/* aux-code.cc: Code-generation-related implementations.		*/

#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>
#include <assert.h>
#include "utils.h"
#include "AST.h"
#include "ClassDecl.h"
#include "CodeContext.h"
#include "CfHeader.h"
#include "CfSource.h"
#include "CtExternal.h"
#include "CtObjectDescriptor.h"
#include "FieldDecl.h"
#include "MethodDecl.h"
#include "aux-code.h"
#include "aux-debug.h"
#include "code.h"
#include "code-assign.h"
#include "code-util.h"
#include "compiler.h"
#include "code-grid.h"
#include "decls.h"
#include "delimit.h"
#include "errors.h"
#include "interface.h"
#include "is-main.h"
#include "jni-bridge.h"
#include "nullify-local.h"
#include "lower.h"
#include "version.h"

/* Set this to true if you want BlockNode::codeGen to void all */
/* its children (kludge).   Put an */
/*     extern bool voidResultsOfBlockStatements; */
/* in your code to get to it. */
bool voidResultsOfBlockStatements = false;

extern void db_print_environ(Environ *);

string javaClassName(Decl *cl)
{
    if (*(cl->container()->name()) == "")
	return *(cl->name());
    else
	return javaClassName(cl->container()) + ("/" + *(cl->name()));
}

bool javaDefinedType(const TreeNode *)
{
    return false;
#if 0
    if (strcmp(t->oper_name(), "TypeNameNode") != 0)
	return false;
    
    Decl *cl = t->decl();
    const char *name = cl->cType().c_str();

    /* "Object" will be handled by titanium */
    if (strcmp(name, "cl_java_lang_Object") == 0)
	return false;

    /* "Exception" will also be handled by titanium */
    if (strcmp(name, "cl_java_lang_Exception") == 0)
	return false;

    /* Other java classes will be interpreted for now */
    if (strncmp(name, "cl_java", 7) == 0)
	return true;
    else
	return false;
#endif
}

string FieldDecl::cFieldName() const
{
  return MANGLE_FIELD(+, *name(), container()->cType());
}

string cFieldName(Decl *decl)
{
  assert( decl->category() == Decl::Field );
  return static_cast< FieldDecl & >( *decl ).cFieldName();
}

Decl *visibleMethod(Decl *meth, ClassDecl *cl)
{
    foriter (decl, cl->environ()->allProperDecls(), EnvironIter) {
	if (decl->category() & Decl::Method) {
	    if ((decl->name() == meth->name()) &&
		(decl->type()->paramTypes()
		 ->typeIdent(meth->type()->paramTypes())))
		return &*decl;
	}
    }
    /* NOT REACHED */
    assert(0);
    return NULL;
}

bool definedEarlier(Decl *meth, ClassDecl *cl)
{
    if ((Decl *) cl == ObjectDecl)
	return false;

    ClassDecl *ancestor = cl->superClass();

    foriter (decl, ancestor->environ()->allProperDecls(), EnvironIter) {
	if (decl->category() & Decl::Method) {
	    if ((decl->name() == meth->name()) &&
		(decl->type()->paramTypes()
		 ->typeIdent(meth->type()->paramTypes())))
		return true;
	}
    }
    return false;
}
				
void emitInstanceMethods(ClassDecl *cl, ClassDecl *outer,
			 void (*outputFun)(MethodDecl &, ostream&), ostream& os)
{
    if (cl) {
	if (cl != ObjectDecl)
	    emitInstanceMethods(cl->superClass(), outer, outputFun, os);
	
	foriter (decl, cl->environ()->allProperDecls(), EnvironIter)
	    if ((decl->category() == Decl::Method) &&
		!(decl->modifiers() & (Common::Static | Common::Final)) &&
		(decl->container() == cl) &&
		(!definedEarlier(&*decl, cl))) {
		Decl *visibleDecl = visibleMethod(&*decl, outer);
		(*outputFun)(static_cast<MethodDecl&>(*visibleDecl), os);
	    }
    }
}

// Emit new instance fields introduced by this class, not inherited
// from an ancestral superclass.
void emitNovelInstanceFields(ClassDecl *cl, FieldPrinter outputFun, CfCode& os)
{
    foriter (decl, cl->environ()->allProperDecls(), EnvironIter)
	if (decl->category() == Decl::Field &&
	    !(decl->modifiers() & Common::Static) &&
	    (decl->container() == cl))
	    (*outputFun)(static_cast<FieldDecl&>(*decl), os);
}

// Emit all instance fields, including those inherited from an
// ancestral superclass.
void emitInstanceFields(ClassDecl *cl, FieldPrinter outputFun, CfCode& os)
{
    if (cl) {
	if (cl != ObjectDecl)
	    emitInstanceFields(cl->superClass(), outputFun, os);

	emitNovelInstanceFields(cl, outputFun, os);
    }
}

void emitStaticMethods(ClassDecl *cl, void (*outputFun)(Decl *, ostream&), ostream& os)
{
    foriter (decl, cl->environ()->allProperDecls(), EnvironIter) {
	if (decl->category() & Decl::Method) {
	    if ((decl->name() == cl->name()) ||
		((decl->modifiers() & Common::Static))) {
		(*outputFun)(&*decl, os);
	    }
	}
    }
}

void emitStaticFields(ClassDecl &cl, FieldPrinter outputFun, CfCode &os)
{
  foriter (decl, cl.environ()->allProperDecls(), EnvironIter)
    if ((decl->category() == Decl::Field)
	&& (decl->modifiers() & Common::Static)) {
      Decl * const container = decl->container();
      if (container == &cl || container->category() != Decl::Interface)
	(*outputFun)(static_cast< FieldDecl & >(*decl), os);
    }
}

void methodDecl(MethodDecl &decl, CfCode &os)
{
    const TypeNode &type = *decl.type();
    const TypeNode &container = *decl.container()->asType();
    const TypeListNode &params = *type.paramTypes();
    const bool isCtor = decl.category() & Decl::Constructor;

    const TypeNode * const thisType = decl.thisType();
    const TypeNode &returnType = isCtor
	? *thisType
	: *decl.type()->returnType();
    
    vector< string > formals;
    if (thisType && !(isCtor && container.isImmutable())) {
	formals.push_back( thisType->cType() );
	os.add( thisType->cType() );
    }
    
    for (int i = 0; i < params.arity(); i++) {
	formals.push_back( params.child(i)->cType() );
	os.add( params.child(i)->cType() );
    }

    os.add( returnType.cType() );
    os << "extern " << returnType.cType()
       << ' ' << decl.cMethodNameStatic()
       << "(" << delimit( formals ) << ");\n";
}

// C Type declaration for a method
string cMethodTypeString(const MethodDecl &method)
{
    TypeListNode *params = method.type()->paramTypes();
    const TypeNode * const thisType = method.thisType();
    vector< string > formals;
    
    if (thisType)
	formals.push_back( thisType->cType() );
    
    for (int i=0; i<params->arity(); i++)
	formals.push_back( params->child(i)->cType() );
    
    return method.type()->returnType()->cType()
	+ "(*)(" + delimit( formals ) + ')';
}

void staticFieldDecl(FieldDecl &fld, CfCode& os)
{
    os.add( fld.cType() );
    os << fld.type()->cType() << " STATIC_DEF(" << fld.cFieldName() << ");\n";
}

static void fieldComp(FieldDecl &fld, CfCode& os)
{
  if (fld.container()->asType()->isImmutable()) {
      os << " && (memcmp(&s1." << fld.cFieldName() << ", &s2."
	 << fld.cFieldName() << ", sizeof("
	 << fld.type()->cType() << ")) == 0)";
  } else {
    os << " && (s1." << fld.cFieldName() << " == s2." << fld.cFieldName() << ")";
  }
}

void externStaticFieldDecl(FieldDecl &fld, CfCode& os)
{
    os << "extern ";
    staticFieldDecl(fld, os);
}

void tableDefn(MethodDecl &method, ostream& os)
{
    if (method.modifiers() & Common::Abstract)
	os << 0;
    else
	os << method.cMethodNameStatic();
    os << ",\n";
}



void emitDefinitions(ClassDecl &cl, CfHeader &hFile, CfSource &cFile )
{
    // For some reason, Object lists Object as its superclass.  :-(
    ClassDecl * const super = (&cl == ObjectDecl ? 0 : cl.superClass());
    
    const string superDescName = super ? super->cDescriptorName() : string("");
    const CtType &superDescType = super
	? static_cast< const CtType & >(super->cDescriptorType())
	: static_cast< const CtType & >(CtObjectDescriptor::classHeader);
    
    const CtType &classDescType = cl.cDescriptorType();
    const string classDescName = cl.cDescriptorName();
    const CtType &className    = cl.cType();
    const string nullifier     = cl.cNullifyLocalName();
    
    const bool isImmutable = cl.asType()->isImmutable();

    // Other classes upon which we depend
    {
	cl.includeSupers( hFile );
	
	cl.includeSelf( cFile );
	cl.includeRequirements( cFile );
    }

    // Instance fields
    className.define( hFile );

    // Class descriptor
    {
	classDescType.define( hFile );
	hFile << "extern " << classDescType << ' ' << classDescName << ";\n";

	classDescType.define( cFile );
	const char *interfaceList = cl.emitInterfaceList( cFile );
	cFile << classDescType << ' ' << classDescName << " = {\n";
	
	if (isImmutable) {
	    cFile << "  Value,\n"
		  << "sizeof( " << className << " ),\n";
	} else {
	    cFile << "  Class,\n"
		  << "  " << interfaceList << ",\n";
	    if (super)
		cFile << "&" << superDescName << ",\n";
	    else
		cFile << "  0,\n";
	    
	    cFile << "sizeof( " << className << " ),\n"
		  << "NULL,\n" // Barrier
		  << nullifier << ",\n";
	    
	    emitInstanceMethods(&cl, &cl, tableDefn, cFile);
	}
	cFile << "};\n\n";
    }

    // Interface Method Table
    if (!isImmutable) {
	hFile << "extern void (*" << cl.cIntfMTName() + "[])();\n\n";

	EmitIntfMT(cl, cFile);
    }

    // Static fields
    {
	emitStaticFields(cl, externStaticFieldDecl, hFile);

	emitStaticFields(cl, staticFieldDecl, cFile);
    }

    // Generate an equals method
    if (isImmutable) {
      hFile << "extern int " << MANGLE_CLASS_EQUALS(<<, className) << "("
	      << cl.asType()->cType() << ", "
	      << cl.asType()->cType() << ");\n";
	cFile << "int " << MANGLE_CLASS_EQUALS(<<, className) << "("
	      << cl.asType()->cType() << " s1, "
	      << cl.asType()->cType() << " s2) {\n"
	      << "return (1";
	emitInstanceFields(&cl, fieldComp, cFile);
	cFile << ");\n"
	      << "}\n";
    }

    // Local field nullifier, used when cloning
    {
	hFile << "void " << nullifier << "(" << className << " *);\n";
	cFile << "void " << nullifier << "(" << className << " *"
	      << MANGLE_THIS_VAR(<<) << ") {\n";
	emitInstanceFields( &cl, fieldNullifyLocal, cFile );
	cFile << "}\n";
    }
}


template< class Iterator >
static void declareAll( CfCode &os, Iterator begin, const Iterator end )
{
    while (begin < end) {
	TreeNode *method = *begin++;
	if (!method->absent()) {
	    MethodDecl &decl = static_cast< MethodDecl & >(*method->decl());
	    assert( decl.category() & (Decl::Constructor | Decl::Method) );	    
	    methodDecl( decl, os );
	}
    }
    os << '\n';
}

template< class Container > inline
static void declareAll( CfCode &os, const Container &container )
{
    declareAll( os, container.begin(), container.end() );
}


void ClassDeclNode::codeGen()
{
//    voidResultsOfBlockStatements = true;

    // reset the temporary counter in ExprNode
    ExprNode::resetTemporary();
    MonitorFetchNode::resetGenerator();

    // reset the temporary count in code-alloc.cc
    extern int initTempNum;
    initTempNum = 0;

    // reset the temporary count in code-foreach.cc
    extern int SRTempNum;
    SRTempNum = 0;

    bool hasNatives = false;
    TypeNode &type = *decl()->asType();

    const CtType &longName = decl()->cType();
    CfHeader hFile( longName );    
    CfSource cFile( longName );

    emitDefinitions(*decl(), hFile, cFile);

    // group members according to where their code goes
    vector< TreeNode * > classInits, fields, methods;
	
    foriter (member, members () -> allChildren (), ChildIter) {
	TreeNode * const node = *member;
	vector< TreeNode * > *category = 0;
	    
	if (isFieldDeclNode(node)) {
	    const FieldDeclNode * const fieldDecl =
		static_cast< const FieldDeclNode * >( node );
	    if (fieldDecl->decl()->modifiers() & Static)
		category = &classInits;
	    else
		category = &fields;
	} else if (isStaticInitNode(node))
	    category = &classInits;
	else if (!node->absent()) {
	    category = &methods;
	    if (node->decl()->modifiers() & Native)
	      hasNatives = true;
	}

	if (category) category->push_back( node );
    }

    // forward declare the string table, currently of unknown size
    cFile.declare( cFile );

    // now put each bit of code in its proper place
    cFile << "static titanium_monitor_t class_monitor;\n";

    cFile << "\nstatic void " << decl()->cInitFieldsName()
	  << "(" << decl()->cType() << " *" << MANGLE_THIS_VAR(<<) << " )\n";
    {
	CodeContext subcontext( cFile );
	EncapsulateInits(fields)->emitStatement(subcontext);
    }
    cFile << '\n';

    hFile << "extern void " << MANGLE_CLASS_INIT(<<, longName) << "();\n";
    cFile << "static void " << MANGLE_CLASS_BUILDST(<<, longName) << "();\n";
    cFile << "void " << MANGLE_CLASS_INIT(<<, longName) << "()\n";
    {
	CodeContext subcontext( cFile );

	subcontext << "monitor_init( &class_monitor );\n"
		   << MANGLE_CLASS_BUILDST(<<, longName) << "();\n";

	EncapsulateInits(classInits)->emitStatement(subcontext);
    }

    if (hasNatives)
	cFile << "\n\n#include \"" << decl()->cNativeName() << "\"\n";

    cFile << '\n';

    declareAll( hFile, methods );
    codeGenAll( cFile, methods );

    cFile << "static void " << MANGLE_CLASS_BUILDST(<<, longName) << "()\n";
    {
      CodeContext context( cFile );
      cFile.internAll( context );
    }
}

const string locComment(TreeNode *t)
{
  static SourcePosn prevPosition;
  const SourcePosn position = t->position();
  
  if (!position.unknown()
      && (position.file != prevPosition.file
	  || position.posn != prevPosition.posn)) {
    prevPosition = position;
    return "/* " + position.asString() + " */ ";
  }
  return "";
}


				/* TreeNode */

void TreeNode::codeGen()
{
}

void TreeNode::codeGen( CfSource & )
{
  undefined( "codeGen( CfSource & )" );
}

void TreeNode::codeGen( CodeContext &context )
{
  foriter (p, allChildren(), ChildIter)
    (*p)->codeGen( context );
}


void CompileUnitNode::codeGen()
{
  if (selectedForCodeGen(false))
    foriter (type, types()->allChildren(), ChildIter)
      if ((*type)->selectedForCodeGen(false))
	(*type)->codeGen();
}

bool CompileUnitNode::selectedForCodeGen(bool force_libs) const
{
  const TreeNode &pack = *package();

  if (pack.absent())
    return true;
  else {
    const string name = pack.decl()->fullName();
    if (name == "ti.internal") return false;
    else if (codeGen_libs || force_libs) return true;
    else {
      const string top = name.substr(0, name.find('.') + 1);
      return top.compare("java.") && top.compare("ti.");
    }
  }
}


bool TemplateDeclNode::selectedForCodeGen( bool ) const
{
  return false;
}


bool ClassDeclNode::selectedForCodeGen(bool) const
{
  if (decl()->container()->fullName() != "ti.domains")
    return true;
  
  const string name = *decl()->name();
  return name != "tiPoint1"
    && name != "tiPoint2"
    && name != "tiPoint3"
    && name != "tiArray1"
    && name != "tiArray2"
    && name != "tiArray3";
}


extern char *currentFilename;  // used for -dumpast switch

void backend()
{
    extern void createDomains_h();
    extern void initPolys();

    initPolys();

    CreateIntfMethods();
    // Generate the .c and .h files for each class
    foreach (f, llist<TreeNode *>, *allFiles) {
      currentFilename = (char *)((*f)->ident()->c_str());
      (*f)->codeGen();
    }

#if 0
    foreach (instance, llist<TreeNode *>, *templateEnv.allInstances())
      {
	const TreeNode &node = **instance;
	Decl &decl = *node.decl();
	TypeNode &type = *decl.asType();
	type.codeGen();
      }
#endif

    if (codeGen_main) {
      // Generate a main.c file that pulls all together
      CfSource mainFile("ti-main");
      mainFile << "#include \"ti-main.h\"\n";

      foreach (f, llist<TreeNode *>, *allFiles) {
	TreeNode *const pack = (*f)->package();
	if (!pack->absent() && pack->decl()->fullName() == "ti.internal")
	  continue;
	    
	// Spit out include directives
	foriter (type, (*f)->types()->allChildren(), TreeNode::ChildIter)
	  if ((strcmp((*type)->oper_name(), "ClassDeclNode") == 0)
	      && (*type)->selectedForCodeGen(false) // filter out tiPoint[123]
	      && !javaDefinedType((*type)->decl()->asType()))
	    (*type)->decl()->includeSelf( mainFile );
      }
      mainFile << "int titaniumCompilerMajorVersion = "
	       << tcMajorVersion << ";" << endl;
      mainFile << "int titaniumCompilerMinorVersion = "
	       << tcMinorVersion << ";" << endl;
      mainFile << "char titaniumCompilerWhatString[] = "
	       << "\"" << tcWhatString << "\";" << endl;
      mainFile << "\n"
	"struct class_header *main_class;\n"
	"\n"
	"int ti_main(int argc, char **argv) {\n";
      /* Debug information */
      emitStackDebugInformation(mainFile);
      mainFile << "  gp_trace_open();\n";

      foreach (f, llist<TreeNode *>, *allFiles) {
	TreeNode *const pack = (*f)->package();
	if (!pack->absent() && pack->decl()->fullName() == "ti.internal")
	  continue;

        // Call class initialization functions
	foriter (type, (*f)->types()->allChildren(), TreeNode::ChildIter) {
	  if ((*type)->selectedForCodeGen(false)
	      && !javaDefinedType((*type)->decl()->asType()))
	    {
	      const ClassDecl &decl = static_cast< const ClassDecl & >(*(*type)->decl());
	      mainFile << "  " << MANGLE_CLASS_INIT(<<, decl.cType()) << "();\n";
	    }
	}
      }

      const MethodDecl &mainMethod = **mainMethods.begin();

      // pause processes after initialization
      mainFile << "  barrier();\n"
	       << "  " << mainMethod.cMethodNameStatic()
	       << "(java_strings_build(argc - 1, argv + 1));\n"
	       << "  barrier();\n"
	       << "  gp_trace_close();\n"
	       << "}\n";

      /* Debug information */
      emitStaticDebugInformation(mainFile, mainMethod);
    }
    
    // Create arrays.h and arrays.c
    instantiateArrays();
    
    // Create domains.h
    createDomains_h();

    // Create type headers
    CtType::elaborateAll();
}
