#include "AST.h"
#include "types.h"
#include "buildexpr.h"

extern const string int2string(int);

map_cstring_to_int *memo = NULL;
int buildexprCount;

void reset_buildexpr()
{
  if (memo != NULL)
    delete memo;
  memo = new map_cstring_to_int;
  buildexprCount = 0;
}

static const string name(int i)
{
  return string("b") + int2string(i);
}

static const string build(const string expr, const CtType &t,
			  CodeContext &context)
{
  int &result = (*memo)[expr];
  if (result == 0) {
    result = ++buildexprCount;
    context.declare(name(result), t);
    context << name(result) << " = " << expr << ';' << endCline;
  }
  return name(result);
}

static const string buildInfix(const string &a, const string &op,
			       const string &b,
			       const CtType &t, CodeContext &context)
{
  return build("(" + a + " " + op + " " + b + ")", t, context);
}

static const string buildPrefix(const string &op, const string &a,
				const CtType &t,
				CodeContext &context)
{
  return build("(" + op + " " + a + ")", t, context);
}

const string product(const string &a, const string &b, CodeContext &context)
{
  return product(a, b, theIntType->cType(), context);
}

const string quotient(const string &a, const string &b, CodeContext &context)
{
  return quotient(a, b, theIntType->cType(), context);
}

const string quotient_with_check(const string &a, const string &b,
				 const string &y, CodeContext &context)
{
  return quotient_with_check(a, b, y, theIntType->cType(), context);
}

const string sum(const string &a, const string &b, CodeContext &context)
{
  return sum(a, b, theIntType->cType(), context);
}

const string difference(const string &a, const string &b, CodeContext &context)
{
  return difference(a, b, theIntType->cType(), context);
}

const string neg(const string &a, CodeContext &context)
{
  return neg(a, theIntType->cType(), context);
}

/////////////////////////////////////////////////////////////////////////////

const string product(const string &a, const string &b, const CtType &t,
		     CodeContext &context)
{
  if (a == "0" || b == "0")
    return "0";
  if (a == "1")
    return b;
  if (b == "1")
    return a;
  return buildInfix(a, string("*"), b, t, context);
}

const string quotient(const string &a, const string &b, const CtType &t,
		      CodeContext &context)
{
  return
    (a == "0" || b == "1") ?
    a :
    buildInfix(a, string("/"), b, t, context);
}

const string quotient_with_check(const string &a, const string &b,
				 const string &y, const CtType &t,
				 CodeContext &context)
{
  if (a == "0" || b == "1")
    return a;
  
  string q = buildInfix(a, string("/"), b, t, context);
  context << "BOUNDS_CHECK((!" << y << " || "
    "(" << q << " * " << b << ") == " << a << "), "
    "\"looping stride is not a multiple of array stride\");" << endCline;
  return q;
}

const string sum(const string &a, const string &b, const CtType &t,
		 CodeContext &context)
{
  return
    (a == "0") ?
    b :
    (b == "0" ? a : buildInfix(a, string("+"), b, t, context));
}

const string difference(const string &a, const string &b, const CtType &t,
			CodeContext &context)
{
  return sum(a, neg(b, t, context), t, context);
}

const string neg(const string &a, const CtType &t, CodeContext &context)
{
  return (a == "0") ? a : buildPrefix(string("-"), a, t, context);
}

