#include "AST.h"
#include "CodeContext.h"
#include "code.h"
#include "lgMacro.h"
#include "decls.h"
#include "code-assign.h"
#include "code-util.h"

const string assignPrefix(AssignKind kind, bool isStatic)
{
  static char *assignMacros[][2] = 
  {
    "",   "",
    "PL", "SPL",
    "PG", "SPG"
  };

  return assignMacros[kind][isStatic];
}

const string assignMacro(AssignKind kind, bool isStatic, bool isLocal)
{
  return assignPrefix(kind, isStatic) + lgMacro("ASSIGN", isLocal);
}

AssignKind assignKind(const TypeNode *type)
{
  if (type->isReference())
    if (type->isLocal()) return assignLocal;
    else return assignGlobal;
  else return assignNormal;
}

const string makeAssign(const string &lvalue, const string &value,
			const bool isLocal, AssignKind kind, bool isStatic)
{
  return assignMacro(kind, isStatic, isLocal)
    + '(' + lvalue + "," +  value + ")";
}


const string TreeNode::receiveAssign(const string &, const string &, AssignKind)
{
  undefined("receiveAssign");
  return "";
}


const string ObjectFieldAccessNode::receiveAssign(const string &lvalue,
						  const string &value,
						  AssignKind kind)
{
  bool isStatic = (decl()->modifiers() & Static) != 0;
  return makeAssign(lvalue, value, isLocalLvalue(), kind, isStatic);
}


const string TypeFieldAccessNode::receiveAssign(const string &lvalue,
						const string &value,
						AssignKind kind)
{
  return makeAssign(lvalue, value, true, kind, true);
}


// True for non-static ObjectFieldAccessNode's whose object is immutable
bool TreeNode::isFieldOfImmutable() const
{
  return false;
}
  
bool FieldAccessNode::isFieldOfImmutable() const
{
  return !(decl()->modifiers() & Static) &&
    accessedObjectType()->isImmutable();
}
  
const string AssignNode::emitExpression2(CodeContext &context)
{
    if (opnd0()->isSimpleLvalue())
      if (isSRArrayAccessNode(opnd1())) {
	const TypeNode * arrayType = opnd1()->array()->type();
	const TypeNode * elemType = arrayType->elementType();
	const bool local = arrayType->isLocal();
	extern string load(const string &dest, const string &address,
			   bool local);
	const string lhs = opnd0()->emitExpression(context);
	const string addr =
	  (static_cast<SRArrayAccessNode *>(opnd1()))->codeString();
	if (local || !elemType->isLocal()) {
	  return load(lhs, addr, local);
	} else {
	  // when accessing a global array of local elements, the
	  // resulting element value must be widened from local to global
	  const string tmp = declareTemporary(context, elemType->cType());
	  context << load(tmp, addr, local) << ';' << endCline;
	  return lhs + " = " +
	    elemType->emitGlobalize(context, "TO_PROC(" + addr + ")", tmp);
	}
      }
      else
	return opnd0()->emitExpression(context) + " = " +
	  opnd1()->emitExpression(context);

    if (opnd0()->isFieldOfImmutable()) {
	const string lval = opnd0()->getLvalue(context);
	const string value = opnd1()->emitExpression(context);
	context << "*(" << lval << ") = " << value << ";" << endCline;
	return value;
    }

    const AssignKind kind = assignKind(type());
    const string lval = opnd0()->getLvalue(context);
    const string value = opnd1()->simpleVar(context);
    context << opnd0()->receiveAssign(lval, value, kind) << ";" << endCline;
    return value;
}
