#ifndef _CODE_DEFS_H_
#define _CODE_DEFS_H_

#include <map>

class DefNode {
public:
  int index;
  TreeNode *t;
  DefNode(TreeNode *tr, int i) : t(tr), index(i) {}
};

// We need a canonical representation of a type that ignores
// modifiers, because a def to a local single foo is treated the same as
// a def to a foo.  That is what xtype is.
typedef string xtype;

typedef map< Decl *, bool, less<Decl *> > map_decl_to_bool;
typedef map< string, llist<DefNode *> *, less<string> > map_string_to_deflist;
typedef map< Decl *, map_string_to_deflist, less<Decl *> >
		map_decl_field_to_deflist;
typedef map< xtype, bool, less<xtype> > map_type_to_bool;
typedef map< xtype, map_string_to_deflist, less<xtype> >
		map_type_field_to_deflist;
typedef map< Decl *, llist<DefNode *> *, less<Decl *> > 
                map_decl_to_deflist;
typedef map< xtype, llist<DefNode *> *, less<xtype> >
		map_type_to_deflist;
typedef map < TreeNode *, llist<DefNode *> *, less<TreeNode *> >
                map_tnode_to_deflist;

class Defs {
public:
  Defs() { unknownMethods = false; nextindex = 0; methodCalls = NULL; }
  ~Defs();

  bool fieldMaybeModified(TypeNode *type, const string &field);
  bool fieldNotModified(TypeNode *type, const string &field);
  bool varModified(Decl *decl);
  bool varMaybeModified(Decl *decl);
  bool arrayMaybeModified(TypeNode *type);
  void merge(TreeNode *t, Bitset *defs, Bitset *kills);
  llist<DefNode *> *getAllDefs(TreeNode *t);
  void reaching(TreeNode *t, Bitset *reaching);
  void connect(TreeNode *nameNode, Bitset *reaching);
  void defAllAliasable(TreeNode *t);
  int size() { return nextindex; };
  inline bool varNotModified(Decl *decl) {
    return !varMaybeModified(decl);
  }

  bool unknownMethods;

  void print(ostream &os, int indent);
  void printVardefs(ostream &os, int indent);
  void printBitset(Bitset *bs, ostream &os, int indent);
  void addCallDefs();

private:
  int nextindex; 
  void addTypeFieldDef(TypeNode *t, const string& field, TreeNode *assignment);

  // type_anyField(type) is whether there is some value for field
  // such that type_field_to_deflist(type, field) is non-null.
  map_type_to_bool _type_anyField;
#define type_anyField(type) \
    _type_anyField[type2xtype(type)]

  // type_field_to_deflist(type, field) is a list of all nodes that modify
  // the given field of an object of the given type.
  map_type_field_to_deflist _type_field_to_deflist;
#define type_field_to_deflist(type, field) \
    _type_field_to_deflist[type2xtype(type)][field]

  // varDefs(decl) is the list of DefNodes that are defs for the given var
  map_decl_to_deflist _varDefs;
#define varDefs(decl) \
    _varDefs[decl]
  
  // arrayDefs(type) is the list of DefNodes that are defs for A[p]
  // for some A of of the given type.
  map_type_to_deflist _arrayDefs;
#define arrayDefs(type) \
    _arrayDefs[type2xtype(type)]

  // callDefs(call) is the list of DefNodes that the given method call defines.
  map_tnode_to_deflist _callDefs;
#define callDefs(call) \
    _callDefs[call]

  inline bool varDefsWithDecl(Decl *decl);
  inline bool fieldDefsWithinType(TypeNode *type);
  inline bool arrayDefsWithinType(TypeNode *type);

public:
  void addVarDef(TreeNode *t, Decl *d);
  void arrayDef(TreeNode *assignment, TypeNode *t);
  void varAssignment(TreeNode *assignment);
  void assignment(TreeNode *t);
  llist <TreeNode *> *methodCalls;
};

#endif
