#include "AST.h"
#include "CodeContext.h"
#include "CtType.h"
#include "UniqueId.h"
#include "decls.h"
#include "jni-bridge.h"
#include "code.h"


const string CodeLiteralExprNode::emitExpression( CodeContext & )
{
  return codeString();
}

const string PrimitiveLitNode::emitExpression2( CodeContext & )
{
    // side-effect free
    return "((" + type()->cType() + ") " + literal().asString() + ")";
}


const string NullPntrNode::emitExpression2( CodeContext & )
{
    // side-effect free
    return "(void *) 0";
}


const string ThisNode::emitExpression2( CodeContext & )
{
    // side-effect free
    return MANGLE_THIS_VAR(+);
}

const string ExprNode::simpleVar( CodeContext &context )
{
  if (cVariable.empty())
    {
      const string var = declareTemporary( context );
      const string val = emitExpression( context );
      context << var << " = " << val << ';' << endCline;
      return var;
    }
  else
    return cVariable;
}

const string ObjectNode::simpleVar( CodeContext &context )
{
    return emitExpression( context );
}

const string ThisNode::simpleVar( CodeContext &context )
{
    return emitExpression( context );
}

bool ObjectNode::isSimpleLvalue()
{
    return true;
}

const string ObjectNode::emitExpression2( CodeContext &context )
{
    // side-effect free
    return name()->decl()->source()->emitUse( context, *this );
}

bool ObjectNode::isLocalLvalue()
{
    return true;
}

bool ObjectNode::isLocalInferredLvalue()
{
    return false;
}

const string ObjectNode::getLvalue( CodeContext &context )
{
    // side-effect free
    return string("&") + name()->decl()->source()->emitUse( context, *this );
}


const string VarDeclNode::emitUse( LocalVars &,
				   const ObjectNode &subject ) const
{
  // side-effect free
  return
    ((decl()->modifiers() & Common::CompilerGenerated) ? 
     MANGLE_TEMP(+, *subject.name()->ident()) :
     MANGLE_STACK_VAR(+, *subject.name()->ident()));
}


const string MethodDeclNode::emitUse( LocalVars &,
				      const ObjectNode &subject ) const
{
    // side-effect free
    return MANGLE_STACK_VAR(+, *subject.name()->ident());
}


const string ConstructorDeclNode::emitUse( LocalVars &,
					   const ObjectNode &subject ) const
{
    // side-effect free
    return MANGLE_STACK_VAR(+, *subject.name()->ident());
}
