#ifndef _TITANIUM_CODE_UTIL_H_
#define _TITANIUM_CODE_UTIL_H_

#include <string>

#include "compiler.h"

class MethodCallNode;
class TreeNode;

void appendWithCommaOperator(string &s, const string &o);
void appendWithSemicolon(string &s, const string &o);

string &unique(void *p);
int &uniqueInt(void *p);
int unique2(void *v, void *w);

Decl * getRectDomainMethodDecl(int arity, const string *name);
Decl * getDomainMethodDecl(int arity, const string *name);
TreeNode *enclosingMethod(const TreeNode *t);
TreeNode *enclosingStmt(const TreeNode *t);
TreeNode *enclosingForeach(const TreeNode *t);
TreeNode *isAncestor(TreeNode *a, TreeNode *b);

// Returns whether s appears in z, a list of n strings.
bool strFind(const char *z[], int n, const string& s);

bool isConstantOrCastOfConstant(const TreeNode *t);
bool isNullOrCastOfNull(const TreeNode *t);
bool isTrivialPointOrDomain(const TreeNode *t);
bool isIntLiteralInRange(const TreeNode *t, int lo, int hi);

bool isPureCallThatCannotFail(TreeNode *t);
bool isSideEffectFreeAndCannotFail(TreeNode *t);

char arithCall(const TreeNode *t);
bool isGridMethod(MethodCallNode *t);
bool isGridMethod_domain(TreeNode *t);
bool isGridMethodMutating(MethodCallNode *t);

bool isTiOrNumericType(const TreeNode *x);
bool isTiOrNumericType(const TypeNode *t);
bool isPointOrIntegralType(const TypeNode *t);

bool isInSystemPackage(Decl *d, string n1, string n2);
bool isInJavaLang(Decl *d);
bool isInJavaIO(Decl *d);

bool isKnownSideEffectFree(TreeNode *md);
bool isKnownPure(TreeNode *md);

bool isStaticInitNode(const TreeNode *t);
bool isFieldDeclNode(const TreeNode *t);
bool isClassDeclNode(const TreeNode *t);
bool isCodeLiteralExprNode(const TreeNode *t);
bool isOFAN(const TreeNode *t);
bool isTypeFAN(const TreeNode *t);
bool isThisFAN(const TreeNode *t);
bool isSFAN(const TreeNode *t);
bool isNameNode(const TreeNode *t);
bool isTypeNameNode(const TreeNode *t);
bool isLabeledStmtNode(const TreeNode *t);
bool isThisNode(const TreeNode *t);
bool isObjectNode(const TreeNode *t);
bool isForEachStmtNode(const TreeNode *t);
bool isForEachPairNode(const TreeNode *t);
bool isMethodCallNode(const TreeNode *t);
bool isMethodCallAssignNode(const TreeNode *t);
bool isPointNode(const TreeNode *t);
bool isDomainNode(const TreeNode *t);
bool isMethodDeclNode(const TreeNode *t);
bool isConstructorDeclNode(const TreeNode *t);
bool isTreeListNode(const TreeNode *t);
bool isCompileUnitNode(const TreeNode *t);
bool isBlockNode(const TreeNode *t);
bool isOmittedNode(const TreeNode *t);
bool isVarDeclNode(const TreeNode *t);
bool isAssignNode(const TreeNode *t);
bool isSynchronizedNode(const TreeNode *t);
bool isParameterNode (const TreeNode *t);
bool isStringLitNode(const TreeNode *t); 
bool isPrimitiveLitNode(const TreeNode *t);
bool isNullPntrNode(const TreeNode *t);
bool isIBroadcastNode(const TreeNode *t);
bool isAllocateNode(const TreeNode *t);
bool isAllocateArrayNode(const TreeNode *t);
bool isExpressionStmtNode(const TreeNode *t);
bool isSwitchBranchNode(const TreeNode *t);
bool isDummyNode(const TreeNode *t);
bool isGotoNode(const TreeNode *t);
bool isBlockNode(const TreeNode *t);
bool isExprNode(const TreeNode *t);
bool isArrayAccessNode(const TreeNode *t);
bool isJavaArrayAccessNode(const TreeNode *t);
bool isPointArrayAccessNode(const TreeNode *t);
bool isTitaniumArrayAccessNode(const TreeNode *t);
bool isSRArrayAccessNode(const TreeNode *t);
bool isOSRArrayAccessNode(const TreeNode *t);
bool isTryStmtNode(const TreeNode *t);
bool isArrayInitNode(const TreeNode *t);
bool isReturnNode(const TreeNode *t);
bool isThisConstructorCallNode(const TreeNode *t);
bool isSuperConstructorCallNode(const TreeNode *t);

bool isPreIncrNode(const TreeNode *t);
bool isPreDecrNode(const TreeNode *t);
bool isPostIncrNode(const TreeNode *t);
bool isPostDecrNode(const TreeNode *t);

bool isBreakNode(const TreeNode *t);
bool isContinueNode(const TreeNode *t);

bool isCastNode(const TreeNode *t);
bool isInstanceOfNode(const TreeNode *t);
bool isDynamicTypeNode(const TreeNode *t);

bool isUnaryPlusNode(const TreeNode *t);
bool isUnaryMinusNode(const TreeNode *t);
bool isUnaryArithNode(const TreeNode *t);

bool isMultNode(const TreeNode *t);
bool isDivNode(const TreeNode *t);
bool isRemNode(const TreeNode *t);
bool isPlusNode(const TreeNode *t);
bool isMinusNode(const TreeNode *t);
bool isBinaryArithNode(const TreeNode *t);

bool isLeftShiftLogNode(const TreeNode *t);
bool isRightShiftLogNode(const TreeNode *t);
bool isRightShiftArithNode(const TreeNode *t);
bool isShiftNode(const TreeNode *t);

bool isEqualityNode(const TreeNode *t);
bool isRelationNode(const TreeNode *t);
bool isBitwiseNode(const TreeNode *t);
bool isLogCondNode(const TreeNode *t);

bool isNotNode(const TreeNode *t);
bool isComplementNode(const TreeNode *t);


bool isFormalParamDecl(const Decl *d);
bool isLocalVarDecl(const Decl *d);
bool isFormalParam(const TreeNode *t);
bool isLocalVar(const TreeNode *t);
bool isConstant(const TreeNode *t);

bool isMultAssignNode(const TreeNode *t);
bool isDivAssignNode(const TreeNode *t);
bool isRemAssignNode(const TreeNode *t);
bool isPlusAssignNode(const TreeNode *t);
bool isMinusAssignNode(const TreeNode *t);
bool isLeftShiftLogAssignNode(const TreeNode *t);
bool isRightShiftLogAssignNode(const TreeNode *t);
bool isRightShiftArithAssignNode(const TreeNode *t);
bool isBitAndAssignNode(const TreeNode *t);
bool isBitXorAssignNode(const TreeNode *t);
bool isBitOrAssignNode(const TreeNode *t);

bool isLazyOptimizeNode(const TreeNode *t);

// Performs the specified action if the -dumpast user argument matches the
// current filename being worked on and the -dumpphase user argument contains
// the specified string. For example:
//
//   DEBUG_PHASE(t->print(cout,0), "opt") 
//
// will dump the AST if -dumpphase contains "opt" and the filename 
// currently being worked on matches the -dumpast user argument.
//
// If -dumpphase is not given, the action will be performed provided the 
// current filename matches.
// If -dumpast is the string "all", all files will have the action
// performed provided the current phase name matches.
// If -dumpphase is given but -dumpast is not given then -dumpast
// defaults to "all".
//
// So on the tc command line, you can say things like:
// tc -dumpast all (all dumps)
// tc -dumpast all -dumpphase cfg (CFG info for all files)
// tc -dumpphase cfg (CFG info for all files)
// tc -dumpast foo.ti (all dumps for foo.ti)
// tc -dumpast foo.ti -dumpphase "cfg lowered" (dump out CFG and lowered AST)
#define DEBUG_PHASE_ENABLED(phasename)					\
  (dump_ast && ((!strcmp(dump_ast, "all") ||				\
		 currentFilename != NULL &&				\
		 (strstr(currentFilename, dump_ast))) &&		\
		(!dump_phase || strstr(dump_phase, phasename))))

#define DEBUG_PHASE(action, phasename)				\
  do if DEBUG_PHASE_ENABLED(phasename) { action; } while (0)


extern bool debug_commonsub;

#define COMPARE(foo, bar)					\
(								\
  (foo)->compare(bar) ?						\
  ((debug_commonsub ? (						\
		       cout << "\tapparently identical:\n",	\
		       (foo)->print(cout, 12),			\
		       cout << "\n\tand:\n",			\
		       (bar)->print(cout, 12),			\
		       cout << '\n',				\
		       0		      ) : 0),		\
   true) :							\
  ((debug_commonsub ? (						\
		       cout << "\tapparently different:\n",	\
		       (foo)->print(cout, 12),			\
		       cout << "\n\tand:\n",			\
		       (bar)->print(cout, 12),			\
		       cout << '\n',				\
		       0		      ) : 0),		\
   false)							\
)

bool compareLists(const llist<TreeNode *> *t, const llist<TreeNode *> *u);

string int2alpha(int i);

#endif /* _TITANIUM_CODE_UTIL_H_ */
