#include "AST.h"
#include "decls.h"
#include "solver.h"


static void handleAccess( const TypeNode *outer,
			  const TypeNode &inner,
			  const TypeNode &result )
{
  if (inner.localInferable())
    {
      inner.localWeaklySubsumedBy( result );
      if (outer && outer->localInferable())
	solver().subsumedBy( *outer, result );
    }
}
			       
      

void ArrayAccessNode::inferLocal( const LocalContext & )
{
  const TypeNode &outer = *array()->type();
  const TypeNode &inner = *outer.elementType();

  handleAccess( &outer, inner, *type() );
}


void ObjectFieldAccessNode::inferLocal( const LocalContext & )
{
  // !! need to go farther out in the case of "object.immutable.field"
  const TypeNode * const outer = (decl()->modifiers() & Common::Static) ? 0 : accessedObjectType();
  const TypeNode &       inner = *decl()->type();

  handleAccess( outer, inner, *type() );
}


void TypeFieldAccessNode::inferLocal( const LocalContext & )
{
  assert( decl()->modifiers() & Static );

  const TypeNode * const outer = 0;
  const TypeNode &       inner = *decl()->type();

  handleAccess( outer, inner, *type() );
}
