#include <cassert>
#include "AST.h"
#include "decls.h"


void Decl::localEnactFromType()
{
  if (type()->isLocal())
    modifiers( (Modifiers) (modifiers() | Common::Local) );
  else
    assert( !(modifiers() & Common::Local) );
}


////////////////////////////////////////////////////////////////////////


void DataDeclNode::localEnact( const LocalEnactContext &context )
{
  if (dtype()->localEnactType( context ))
    localReportChange( decl()->errorName(), *dtype() );
}


void FieldDeclNode::localEnact( const LocalEnactContext &context )
{
  DataDeclNode::localEnact( context );
  decl()->localEnactFromType();
}


void MethodNode::localEnact( const LocalEnactContext &context )
{
  TypeNode &type = *decl()->type();
  const bool thisReference =
    !(decl()->modifiers() & Static) &&
    decl()->container()->asType()->localInferable();

  if (thisReference && type.localEnactType( context ))
    {
      decl()->localEnactFromType();
      localReportChange( "dispatch" );
    }
  
  if (type.returnType()->localEnactType( context ))
    {
      localReportChange( "return" );
      assert( !(decl()->modifiers() & Native) );
    }
}


void ParameterNode::localEnact( const LocalEnactContext &context )
{
  if (dtype()->localEnactType( context ))
    {
      localReportChange( decl()->errorName(), *dtype() );
      assert( !(parent()->parent()->decl()->modifiers() & Native) );
    }
}
