#include <utility>
#include "AST.h"
#include "compiler.h"
#include "context.h"
#include "enact-context.h"
#include "options.h"
#include "solver.h"
#include "st-field.h"
#include "stats.h"
#include "template.h"
#include "typecheck.h"
#include "using-std.h"
#include "utils.h"


void inferLocal()
{
  if (opt_local) {
    llist< TreeNode * > * const allInsts = templateEnv.allInstances();
    llist< TreeNode * > * const allTrees = extend( allInsts, allFiles );
    
    LocalSolver solver;
    foreach( tree, llist< TreeNode * >, *allTrees )
      (*tree)->unshareTypeTree();

    StaleSet stale;
    LocalContext context( stale );
    foreach( tree, llist< TreeNode * >, *allTrees )
      (*tree)->inferLocalTree( context );

    LocalSolution solution( solver.solution() );
    foreach( tree, llist< TreeNode * >, *allTrees )
      (*tree)->localEnactTree( solution );

    foreach_const (call, StaleSet, stale)
      (*call)->localRebind();

    // sanity checking
    TreeNode::TypeContext typeContext;
    foreach( tree, llist< TreeNode * >, *allFiles )
      (*tree)->typecheck( &typeContext );
    
    if (debug_lqi)
      LocalEnactContext::printStats( cerr );
  }
}
