#include <cstdio>
#include <iostream.h>
#include "AST.h"
#include "code-call.h"
#include "code-foreach.h"
#include "code-ipanal.h"
#include "code.h"
#include "compiler.h"
#include "config.h"
#include "decls.h"
#include "errors.h"
#include "local/options.h"
#include "optimize.h"
#include "postpone.h"
#include "streq.h"
#include "stats.h"
#include "templates/depth.h"
#include "templates/instantiate.h"
#include "utils.h"

int DEBUG = 0;
bool alias_anal = false;
bool compilerLocComments = true;
bool debug_opt = false;
bool debug_commonsub;
bool debug_defs = false;

// Optimization is currently off by default, but later will be on by default.
// CM: Actually, leave optimization off here. The wrapper script will handle
// the user's view of the default.
bool opt_osr = false;
bool opt_dce = false;
bool opt_sr = false;
bool opt_lift = false;

bool test_sr = false;
bool codeGen_libs = true;
bool codeGen_main = true;
bool use_sglobal_inference = true;
bool use_goto = false;
bool use_nonTrivialSizeCheck = false;
bool force_java = false, force_ti = false;
extern bool sizeofRuntimeIntIs64;
extern bool sizeofRuntimeCharIs32;
extern bool sizeofRuntimeShortIs32;
bool parse_only = false;
string codeGen_outdir = "./";
int debugger_level = 0;
bool line_directives = true; /* should be true in production */
char *dump_ast, *dump_phase;

main(int argc, char* argv[])
{
  bool frontend_only = false;
  int args_terminated = 0;

  while (argc > 1 && argv[1][0] == '-' && !args_terminated) {
    if (strcmp(argv[1], "-fonly") == 0) {
      frontend_only = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-parse") == 0) {
      parse_only = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-g0") == 0) {
      debugger_level = 0;
      argc--;
      argv++;
    }
    else if ((strcmp(argv[1], "-g") == 0) || (strcmp(argv[1], "-g1") == 0)) {
      debugger_level = 1;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-noline") == 0) {
      line_directives = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-line") == 0) {
      line_directives = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-debug") == 0) {
      DEBUG = 1;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-sizeof") == 0) {
      argc--;
      argv++;
      if (argc > 1) {
	if (strcmp(argv[1], "jint=64") == 0) {
	  sizeofRuntimeIntIs64 = true;
	  argc--;
	  argv++;
	}
	else if (strcmp(argv[1], "jint=32") == 0) {
	  sizeofRuntimeIntIs64 = false;
	  argc--;
	  argv++;
	}
	else if (strcmp(argv[1], "jchar=32") == 0) {
	  sizeofRuntimeCharIs32 = true;
	  argc--;
	  argv++;
	}
	else if (strcmp(argv[1], "jchar=16") == 0) {
	  sizeofRuntimeCharIs32 = false;
	  argc--;
	  argv++;
	}
	else if (strcmp(argv[1], "jshort=32") == 0) {
	  sizeofRuntimeShortIs32 = true;
	  argc--;
	  argv++;
	}
	else if (strcmp(argv[1], "jshort=16") == 0) {
	  sizeofRuntimeShortIs32 = false;
	  argc--;
	  argv++;
	}
      }
    }
    else if (strcmp(argv[1], "-trivialSizeCheck") == 0) {
      use_nonTrivialSizeCheck = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-notrivialSizeCheck") == 0) {
      use_nonTrivialSizeCheck = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-goto") == 0) {
      use_goto = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-nogoto") == 0) {
      use_goto = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-debugopt") == 0) {
      debug_opt = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-debugdefs") == 0) {
      debug_defs = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-srtest") == 0) {
      test_sr = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-stats") == 0) {
      stat_foreach = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-O0") == 0) {
      opt_dce = false;
      opt_osr = false;
      opt_sr = false;
      opt_lift = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-O") == 0 || strcmp(argv[1], "-O1") == 0) {
      opt_sr = true;
      opt_lift = true;
      opt_osr = true;
      opt_dce = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-OSR") == 0) {
      opt_sr = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-OnoSR") == 0) {
      opt_sr = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-OOSR") == 0) {
      opt_osr = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-OnoOSR") == 0) {
      opt_osr = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-ODCE") == 0) {
      opt_dce = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-OnoDCE") == 0) {
      opt_dce = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-Olift") == 0) {
      opt_lift = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-Onolift") == 0) {
      opt_lift = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-Ofinalize") == 0) {
      opt_finalize = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-Onofinalize") == 0) {
      opt_finalize = false;
      argc--;
      argv++;
    }
#ifdef USE_LOCAL_INFERENCE
    else if (strcmp(argv[1], "-Olocal") == 0) {
      opt_local = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-Onolocal") == 0) {
      opt_local = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-ibane") == 0) {
      opt_ibane = argv[2];
      argc -= 2;
      argv += 2;
    }
    else if (strcmp(argv[1], "-debuglqi") == 0) {
      debug_lqi = true;
      argc--;
      argv++;
    }
#endif // USE_LOCAL_INFERENCE
#ifdef USE_SGLOBAL_INFERENCE
    else if (strcmp(argv[1], "-infer-sglobal") == 0) {
	use_sglobal_inference = true;
	argc--;
	argv++;
    }
#endif // USE_SGLOBAL_INFERENCE
    else if (strcmp(argv[1], "-noinfer-sglobal") == 0) {
	use_sglobal_inference = false;
	argc--;
	argv++;
    }
    else if (strcmp(argv[1], "-nolibs") == 0) {
      codeGen_libs = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-libs") == 0) {
      codeGen_libs = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-nomain") == 0) {
      codeGen_main = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-main") == 0) {
      codeGen_main = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-outdir") == 0) {
      codeGen_outdir = argv[2];
      codeGen_outdir += '/';
      argc -= 2;
      argv += 2;
    }
    else if (strcmp(argv[1], "-dumpast") == 0) {
      dump_ast = argv[2];
      argc -= 2;
      argv += 2;
    }
    else if (strcmp(argv[1], "-dumpphase") == 0) {
      dump_phase = argv[2];
      if (dump_ast == NULL)
	dump_ast = "all";
      argc -= 2;
      argv += 2;
    }
    else if (strcmp(argv[1], "-java") == 0) {
      force_java = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-ti") == 0) {
      force_ti = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-AA") == 0) {
      alias_anal = true;
      argc--;
      argv++;
    }        
    else if (strcmp(argv[1], "-noAA") == 0) {
      alias_anal = false;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "--") == 0) {
      args_terminated = 1;
      argc--;
      argv++;
    }
    else if (streq(argv[1], "-template-depth")) {
      char *end = 0;
      int value = strtol( argv[2], &end, 0 );
      argc--;
      argv++;
      if (*end)
	cerr << "Ignoring garbled template depth: " << argv[2] << endl;
      else
	maxTemplateDepth = value;
    }
    else if (strcmp(argv[1], "-debug-instantiation") == 0) {
      debug_instantiation = true;
      argc--;
      argv++;
    }
    else if (strcmp(argv[1], "-debug-postpone") == 0) {
      debug_postpone = true;
      argc--;
      argv++;
    }
    else {
      cerr << "Ignoring unknown option: " << argv[1] << '\n';
      argc--;
      argv++;
    }      
  }

  // If testing strength reduction, make sure we do that optimization.
  if (test_sr) opt_sr = true;

  debug_commonsub = false; 
  // debug_commonsub = debug_opt;

  initTheSystem();

  if (argc > 1)
    for (int source = 1; source < argc; ++source)
      load(argv[source], !force_java);
           //!force_java && (force_ti || isTitanium(argv[source])));
  else
    load("<stdin>", stdin, !force_java);

  if (!parse_only)
    {
      staticSemantics();

      if (!frontend_only && NumErrors() == 0) {
	// temporary hack---to be removed later! 
	if (!alias_anal && getenv("MAIN"))
	  setupBblocks();

	/* Interprocedural analyses (currently only AA). */
	if (alias_anal) {
	  /* Set up all basic blocks first. */
	  setupBblocks();       
	  IPAnal();
	}

	/* Interprocedural opts can go here. */

	/* Mark method bodies etc. for optimization. */
	lazyOptimize();

	/* Generate code, doing some optimization as we go. */
	backend();
      }
    }

  exit(NumErrors() > 0 ? 1 : 0);
}
