#include <pthread.h>
#include <stdlib.h>
#include "gc.h"			/* replaces certain pthread calls */
#include "main.h"
#include "procs.h"
#include "spawn.h"


static int ti_argc;
static char **ti_argv;

static pthread_t *threads;

#ifdef NO_PROCID_REGISTER
pthread_key_t myProcKey;
#endif /* NO_PROCID_REGISTER */


static void *start( void *id )
{
#ifdef NO_PROCID_REGISTER
  pthread_setspecific( myProcKey, id );
#else  /* process id in register */
  MYPROC = (int) id;
#endif /* process id in register */
  
  ti_main( ti_argc, ti_argv );
  return 0;
}


static void create()
{
  int count;
  pthread_attr_t attr;
  pthread_attr_init( &attr );
  /* Explicitly calling pthread_attr_setscope may causes systems that do
     not fully implement pthread to fail in the create.  So, just use the
     default value. */
  /* pthread_attr_setscope( &attr, PTHREAD_SCOPE_SYSTEM ); */

  threads = ti_malloc( PROCS * sizeof( pthread_t ) );

  for (count = PROCS; count--; )
    pthread_create( &threads[ count ], &attr, start, (void *) count );
}


static void join()
{
  int count;
  
  for (count = PROCS; count--; )
    pthread_join( threads[ count ], 0 );

  ti_free( threads );
}


void spawn( int argc, char *argv[] )
{
#ifdef NO_PROCID_REGISTER
  pthread_key_create( &myProcKey, 0 );
#endif /* NO_PROCID_REGISTER */
  
  ti_argc = argc;
  ti_argv = argv;
  
  if (PROCS > 1)
    {
      create();
      join();
    }
  else
    start( 0 );
}
