/* ti_array.c */
/* This file assumes T is a type and N is an arity. */

/* To add a method you must (as of 12/97):
     edit code-grid.cc, adding a line to the part that looks like
        D("DOMAIN_METHOD", "_ti_arraymethod_domain(" << t << ", " << n << ")",
     edit tlib/ti/internal/tiArray.java    
     edit titanium.h
     add the method defn to this file or ti_array.h
     */

#include "config.h"
#include "runtime-options.h"
#include "backend.h"
#include "comm.h"
#include "assertExists.h"

#include <gc.h>
#include <primitives.h>
#include <regions.h>
#include <ti-memory.h>

#if (N == 2) || (N == 3)
#include <slice.h>
#endif

#define N_LARGER_THAN_1 N_MINUS_1
#define N_IS_1 (N && !N_LARGER_THAN_1)

#ifndef GLOBAL_ARRAY
#define GLOBAL_ARRAY 1
#endif

#ifndef DEBUG_ARRAYS
#define DEBUG_ARRAYS 0
#endif

#ifndef ASSERT_SAME_BYTES
#define ASSERT_SAME_BYTES
static void assertSameBytes(char *x, char *y, int len)
{
  while (len-- > 0) assert(*x++ == *y++);
}
#endif /* ASSERT_SAME_BYTES */

#ifndef DUMPABLE_ARRAYS
#define DUMPABLE_ARRAYS 0
#endif

#ifndef DEFAULT_DUMP_DEST
#define DEFAULT_DUMP_DEST "|/disks/orodruin/titanium/work/pike/bin/aview"
#endif

#if GLOBAL_ARRAY
#define deref DEREF_GLOBAL
#define INDEX INDEX_GLOBAL
#define LOCAL_PART TO_LOCAL
#define PROC_PART  TO_PROC
#else
#define deref DEREF_LOCAL
#define INDEX INDEX_LOCAL
#define LOCAL_PART
#define PROC_PART(p) MYPROC
#endif

#if GLOBAL_ARRAY && !GLOBAL_ELEMENTS
# define assignable 0
#else
# define assignable 1
#endif

#if ELEMENTS_ARE_ATOMIC
# define array_ralloc ralloc_atomic
# define array_malloc ti_malloc_atomic_huge
#else
# define array_ralloc ralloc
# define array_malloc ti_malloc_huge
#endif

#define PRINT_POINT(p, f) 					\
do {								\
  jint _i;							\
  fprintf(f, "[");						\
  for (_i = 1; _i <= N; _i++)					\
    fprintf(f, (_i > 1) ? ", %d" : "%d", POINT_INDEX(p, _i));	\
  fprintf(f, "]");						\
} while (0)

#define PRINT_DOMAIN(R, f)			\
do {						\
  jint _i;					\
  ti_POINT p = RECTDOMAIN_STRIDE(R);		\
  fputs("[", f);				\
  PRINT_POINT(RECTDOMAIN_MIN(R), f);		\
  fputs(" : ", f);				\
  PRINT_POINT(RECTDOMAIN_UPB(R), f);		\
  for (_i = 1; _i <= N; _i++)			\
    if (POINT_INDEX(p, _i) >= 2) break;		\
  if (_i <= N) {				\
    fputs(" : ", f);				\
    PRINT_POINT(p, f);				\
  }						\
  fputs("]", f);				\
} while (0)


#if DUMPABLE_ARRAYS
void ti_dump(ti_ARRAY *px, char *dest, char *userArg)
{
  ti_ARRAY x = *px;
  FILE *f;
  ti_RECTDOMAIN R = x.domain;
  ti_POINT p, q, s;
  jint lo[N], hi[N], stride[N], i;

  if (dest == NULL || dest[0] == '\0') dest = DEFAULT_DUMP_DEST;
  
  if (dest[0] == '|')
    f = popen(dest + 1, "w");
  else
    f = fopen(dest, "w");

  if (f == NULL) {
    fprintf(stderr, "Unable to open `%s'; dump request failed.\n", dest);
    return;
  }

  if (userArg != NULL) fprintf(f, "%s\n", userArg);
  fprintf(f, "%d\n", N);
  PRINT_DOMAIN(x.domain, f);
  fprintf(f, "\ndomainaddr <%d, %x>", TO_PROC(x.domain), TO_LOCAL(x.domain));
  fprintf(f, "\nstride");
  for (i = 0; i < N; i++) fprintf(f, " %d", x.stride[i]);
  fprintf(f, "\nbase");
  for (i = 0; i < N; i++) fprintf(f, " %d", x.base[i]);
  fprintf(f, "\nsideFactors");
  for (i = 0; i < N; i++) fprintf(f, " %d", x.sideFactors[i]);
  fprintf(f, "\ndesc %lx\nA %lx\nancestor %lx\nwhere %s\n",
	  (long)px, (long)LOCAL_PART(x.A), (long)LOCAL_PART(x.ancestor),
	  x.where);
  fprintf(f, "elementSize %d\n", sizeof(T));
  fputs("elementType " T_string "\n", f);

  p = RECTDOMAIN_MIN(R);
  q = RECTDOMAIN_MAX(R);
  s = RECTDOMAIN_STRIDE(R);
  for (i = 0; i < N; i++) {
    lo[i] = POINT_INDEX(p, i + 1);
    hi[i] = POINT_INDEX(q, i + 1);
    if ((stride[i] = POINT_INDEX(s, i + 1)) == 0) stride[i] = 1;
  }
  
  forall(e, x, lo, hi, stride) {
    fputc('!', f);
    PRINT_FORALL_POINT(f);
    fputc(',', f);
    if (strcmp("jint", T_string) == 0)
      fprintf(f, "%d", deref(e));
    else if (strcmp("jdouble", T_string) == 0)
      fprintf(f, "%g", deref(e));
    else if (strcmp("jfloat", T_string) == 0)
      fprintf(f, "%f", deref(e));
    else {
      T buf = deref(e);
      for (i = 0; i < sizeof(T); i++)
	fprintf(f, "%02x", (unsigned int)(((unsigned char *) &buf)[i]));
    }
    fputc('!', f);
  }

  if (dest[0] == '|')
    pclose(f);
  else
    fclose(f);
}
#endif

jboolean ti_isnull(ti_ARRAY x)
{
  return (LOCAL_PART(x.A) == 0);
}

ti_ARRAY ti_empty()
{
  ti_ARRAY x;
  memset(&x, 0, sizeof(x));
  return x;
}

ti_ARRAY ti_construct(const char *where, Region pregion, ti_RECTDOMAIN R)
{
  ti_ARRAY r;
  jint i, max[N], size;
  ti_POINT p;
  T *array;

  r.domain = R;
  r.where = where;
#if DUMPABLE_ARRAYS
  r.dumpfn = (void(*)()) ti_dump;
#endif
  
  p = RECTDOMAIN_MIN(R); /* minimum point in R */
  for (i = 0; i < N; i++)
    r.base[i] = POINT_INDEX(p, i + 1);
  
  p = RECTDOMAIN_STRIDE(R); /* stride of R */
  for (i = 0; i < N; i++) 
    if ((r.stride[i] = POINT_INDEX(p, i + 1)) == 0) r.stride[i] = 1;
  
  p = RECTDOMAIN_MAX(R); /* maximum point in R */
  for (i = 0; i < N; i++)
    max[i] = (POINT_INDEX(p, i + 1) - r.base[i]) / r.stride[i];
    
  r.sideFactors[N - 1] = 1;
  if (N > 1) {
    jint sf = 1;
    for (i = N - 2; i >= 0; i--)
      r.sideFactors[i] = sf *= (1 + max[i + 1]);
  }

  size = 1 + max[0];
  if (N > 1)
    size *= r.sideFactors[0];

#if DEBUG_ARRAYS
  fprintf(stderr, "p%d: new array with domain <%d, %x> = ",
	  MYPROC, TO_PROC(R), (int) TO_LOCAL(R));
  PRINT_DOMAIN(R, stderr);
  fprintf(stderr, "\nArray<%d>:\tbase\tmax\tstride\tsideFactors\n", N);
  for (i = 0; i < N; i++) {
    fprintf(stderr, " %d\t\t%d\t%d\t%dt%d\\n", i,
	    r.base[i], max[i], r.stride[i], r.sideFactors[i]);
  }
#endif

  if (pregion)
    array = (T *)array_ralloc(pregion, size * sizeof(T));
  else
    array = (T *)array_malloc(size * sizeof(T));

  memset(array, 0, size * sizeof(T));

#if GLOBAL_ARRAY
  TO_GLOBAL(r.A, MYPROC, array);
#else
  r.A = array;
#endif
  r.ancestor = r.A;

#if EXPLICITLY_STORE_CREATOR
  r.creator = MYPROC;
#endif

  return r;
}

jint ti_conv(const char *where, const ti_ARRAY *x, ti_POINT p)
{
  jint sum = 0, q, qm, s;
#if N_LARGER_THAN_1
  jint i;
#else
#define i 1
#endif

  assertExists(*x, where);
#if N_LARGER_THAN_1
  for (i = N; i >= 1; i--)
#endif
    {
      q = POINT_INDEX(p, i);
      qm = q - x->base[i-1];
#if BOUNDS_CHECKING
      {
	jint lo = POINT_INDEX(RECTDOMAIN_MIN(x->domain), i);
	jint hi = POINT_INDEX(RECTDOMAIN_MAX(x->domain), i);
	s = POINT_INDEX(RECTDOMAIN_STRIDE(x->domain), i);
	if (s > 1 && ((q - lo) % s) != 0) {
	fprintf(stderr, "bounds checking error at %s:\n  failed condition: [",
		where);
	PRINT_POINT(RECTDOMAIN_MIN(x->domain), stderr);
	fprintf(stderr, " : ");
	PRINT_POINT(RECTDOMAIN_UPB(x->domain), stderr);
	fprintf(stderr, " : ");
	PRINT_POINT(RECTDOMAIN_STRIDE(x->domain), stderr);
	fprintf(stderr, "].contains(");
	PRINT_POINT(p, stderr);
	fprintf(stderr, ")\n  array was allocated at %s\n", x->where);
	assert(("array index out of bounds", 0));
	exit(-2);
      }
	else if (q < lo || q > hi) {
	fprintf(stderr, "bounds checking error at %s:\n  failed condition: ",
		where);
	PRINT_POINT(RECTDOMAIN_MIN(x->domain), stderr);
	fprintf(stderr, " <= ");
	PRINT_POINT(p, stderr);
	fprintf(stderr, " <= ");
	PRINT_POINT(RECTDOMAIN_MAX(x->domain), stderr);
	fprintf(stderr, "\n  array was allocated at %s\n", x->where);
	assert(("array index out of bounds", 0));
	exit(-2);
      }
  	if ((qm % x->stride[i-1]) != 0)
	  assert(("internal error: bad array descriptor", 0));
      }
#endif /* BOUNDS_CHECKING */
      /* At this point, x->stride[i-1] should divide qm.  Of course, we only
	 know for sure if bounds checking is on. */
      sum += (qm / x->stride[i-1]) * x->sideFactors[i-1];
    }
#if !N_LARGER_THAN_1
#undef i  
#endif

#if 0
  fprintf(stderr, "conv(");
  PRINT_POINT(p, stderr);
  fprintf(stderr, ") = %d\n", sum);
#endif
  return sum;
}



#if N_LARGER_THAN_1
ti_ARRAY_SLICE SLICE_METHOD(const ti_ARRAY x, jint k, jint j)
{
  jint i, l;
  ti_ARRAY_SLICE r;
  ti_POINT p;

  assertExists(x, "(in slice method)");
  r.domain = RECTDOMAIN_SLICE(x.domain, k);
  r.where = x.where;

  for (i = l = 0; i < N_MINUS_1; i++, l++) {
    if (i == k - 1) ++l;
    r.stride[i] = x.stride[l];
    r.base[i] = x.base[l];
    r.sideFactors[i] = x.sideFactors[l];
  }    

#if BOUNDS_CHECKING
  p = POINT_SET(RECTDOMAIN_MIN(x.domain), k - 1, j);
  ti_conv("(in slice method)", &x, p); /* just for bounds check */
#endif
  INDEX(r.A, x.A, (j - x.base[k - 1]) / x.stride[k - 1] * x.sideFactors[k - 1]);
  
  r.ancestor = x.ancestor;
#if EXPLICITLY_STORE_CREATOR 
  r.creator = x.creator;
#endif

#if DEBUG_ARRAYS
  fprintf(stderr, "slice(%d, %d) of [%x]\n  Array<%d>:\tbase\tstride\tsideFactors\n", k, j, (int) LOCAL_PART(x.A), N);
  for (i = 0; i < N; i++) {
    fprintf(stderr, "   %d\t\t%d\t%d\t%d\n", i,
	    x.base[i], x.stride[i], x.sideFactors[i]);
  }
  fprintf(stderr, "is [%x]\n  Array<%d>:\tbase\tstride\tsideFactors\n",
	  (int) LOCAL_PART(r.A), N - 1);
  for (i = 0; i < N - 1; i++) {
    fprintf(stderr, "   %d\t\t%d\t%d\t%d\n", i,
	    r.base[i], r.stride[i], r.sideFactors[i]);
  }
#endif

  return r;
}
#else
void SLICE_METHOD(const ti_ARRAY x, jint k, jint j)
{
  fputs("Error: Slice on 1d array (ignored)\n", stderr);
}
#endif

#if assignable
/* x.exchange(w) is equivalent to:
   ordered_foreach (p in Ti.myTeam().domain()) x[p] = broadcast w from p[1];

   The current implementation assumes myTeam().domain() is [0 : PROCS - 1].
   */
void EXCHANGE_METHOD(const ti_ARRAY x, T w)
{
  assertExists(x, "(in exchange method)");
  if (N != 1) fprintf(stderr, "Error: exchange() used on %dd array; "
		      "it should only be used on 1d arrays.\n", N);
#if N_IS_1
  {
    jint p1;

    for (p1 = 0; p1 < PROCS; p1++) {
      T receive;
      {
	BROADCAST_BEGIN( T, p1 );	
	if (p1 == MYPROC)
	  BROADCAST_SEND(    receive, T, w );
	else
	  BROADCAST_RECEIVE( receive, T, p1 );
	BROADCAST_END( receive, T, p1 );
      }
      {
	ti_POINT p;
	PTR_TO_T target;
	mi9constructmT8tiPoint17domains2ti(p, p1);
	INDEX(target, x.A, ti_conv("(in exchange method)", &x, p));
	assign(target, receive);
      }
    }
  }
#endif
  /*
   * Ensure that every process is done writing into its respective
   * array before continuing.  Otherwise, one process can see
   * another process's array in a partially-written state.
   */
  barrier();
}

/* This is a "copy descriptor", which encapsulates the domain and array 
   information a remote node will need to pack array data into a buffer. */

typedef struct {
  ti_ARRAY x;      /* Array involved in the copying. */
  ti_RECTDOMAIN R; /* The subarray that will actually be copied. */
} copy_desc_t;     /* Note that this name is actually a macro, to preserve
		      uniqueness among different array types. */

/* Takes the given copy descriptor and returns a buffer that contains the
   desired elements of the array. The array_data_size argument is an output
   parameter. If the buf argument is not null, the data will be copied into 
   it; otherwise memory will be allocated. */
T *PACK_METHOD(void *copy_desc_ptr, int *array_data_size, void *allocated) {

  ti_ARRAY x;
  ti_RECTDOMAIN R;
  int i, lo[N], hi[N], stride[N], size, idx;
  ti_POINT p, q, s;
  T *buf;
  copy_desc_t copy_desc = *(copy_desc_t *)copy_desc_ptr;
  
  R = copy_desc.R;
  x = copy_desc.x;

  if (RECTDOMAIN_ISNULL(R)) return;
  p = RECTDOMAIN_MIN(R);
  q = RECTDOMAIN_MAX(R);
  s = RECTDOMAIN_STRIDE(R);
  for (i = 0; i < N; i++) {
    lo[i] = POINT_INDEX(p, i + 1);
    hi[i] = POINT_INDEX(q, i + 1);
    if ((stride[i] = POINT_INDEX(s, i + 1)) == 0) stride[i] = 1;
  }
  size = RECTDOMAIN_NUMPOINTS(R) * sizeof(T);
  if (allocated) {
    buf = allocated;
  }
  else {
    buf = (T *)malloc(size);
  }
  idx = 0;
  forall(e, x, lo, hi, stride, deref(buf[idx++],e););
  *array_data_size = size;
  return buf;
}

/* Unpacks the given buffer into the array with the given copy
   descriptor. */
void UNPACK_METHOD(copy_desc_t *copy_desc_ptr, T *buf) {
  int i, lo[N], hi[N], stride[N], size, idx;
  ti_POINT p, q, s;
  ti_ARRAY x = copy_desc_ptr->x;
  ti_RECTDOMAIN R = copy_desc_ptr->R;
  T tmp;

  assertExists(x, "(in unpack method)");
  
  if (RECTDOMAIN_ISNULL(R)) return;
  p = RECTDOMAIN_MIN(R);
  q = RECTDOMAIN_MAX(R);
  s = RECTDOMAIN_STRIDE(R);
  for (i = 0; i < N; i++) {
    lo[i] = POINT_INDEX(p, i + 1);
    hi[i] = POINT_INDEX(q, i + 1);
    if ((stride[i] = POINT_INDEX(s, i + 1)) == 0) stride[i] = 1;
  }
  idx = 0;
  forall(e, x, lo, hi, stride, weak_assign(e, buf[idx++]););
  ti_sync();
}

void COPY_METHOD(const ti_ARRAY x, const ti_ARRAY y)
{
  ti_RECTDOMAIN R;

  assertExists(x, "(in copy method)");
  assertExists(y, "(in copy method)");

  R = RECTDOMAIN_INTERSECTION(y.domain, x.domain);
  if (RECTDOMAIN_ISNULL(R)) return;

#if GLOBAL_ARRAY
#if defined(MEMORY_DISTRIBUTED) && !(defined(USE_RUNTIME_CRAY_GNU_C) || defined(USE_RUNTIME_CRAY_T3E_C))
  {
    copy_desc_t copy_desc;
    T *buf;

    int toProc = TO_PROC(x.A);
    int fromProc = TO_PROC(y.A);

    copy_desc.R = R;

#ifdef CLUMP_BACKEND
    if ((NODE_MAP(toProc) == MYNODE) && (NODE_MAP(fromProc) != MYNODE)) {
#else 
    if ((toProc == MYPROC) && (fromProc != MYPROC)) {
#endif
      int num_bytes = RECTDOMAIN_NUMPOINTS(R) * sizeof(T);

      copy_desc.x = y;
       
      buf = (T *)malloc(num_bytes);

      get_array((void *)PACK_METHOD, &copy_desc, sizeof(copy_desc), 
		fromProc, (void *)buf);       
      copy_desc.x = x;
      UNPACK_METHOD(&copy_desc, buf);
      free(buf);
      ti_sync();
      return;
    }
#ifdef CLUMP_BACKEND
    else if ((NODE_MAP(toProc) != MYNODE) && (NODE_MAP(fromProc) == MYNODE)) {
#else
    else if ((toProc != MYPROC) && (fromProc == MYPROC)) {
#endif
      int num_bytes;
      int i;

      copy_desc.x = y;
      buf = PACK_METHOD(&copy_desc, &num_bytes, NULL);
      copy_desc.x = x;
      put_array((void *)UNPACK_METHOD, &copy_desc, sizeof(copy_desc),
		buf, num_bytes, toProc);
      free(buf);
      ti_sync();
      return;
    }
  }
#else
  {
    /* If the RectDomains and sideFactors are the same, do a bulk copy. */
    if (RECTDOMAIN_EQUAL(y.domain, x.domain)) {
      int i;
      int found_diff = 0;
      for (i = 0; i < N ; i++) {
	found_diff = found_diff || (x.sideFactors[i] != y.sideFactors[i]);
      }
      if (!found_diff) {
	/* RectDomains and sideFactors are the same. */
	size_t size;

	/* assert(RECTDOMAIN_NUMPOINTS(x.domain) ==
	   RECTDOMAIN_NUMPOINTS(y.domain)); */
	size = sizeof(T) * RECTDOMAIN_NUMPOINTS(y.domain);
	/* Sets e and f to point to the start of the area of interest on each
	   array. */

	if ((TO_PROC(x.A) == MYPROC) && (TO_PROC(y.A) == MYPROC)) {
	  /* local<-local */
	  memcpy(TO_LOCAL(x.A), TO_LOCAL(y.A), size);
	} else if (TO_PROC(x.A) == MYPROC) {
	  /* local<-remote */
	  ti_bulk_read(TO_LOCAL(x.A), TO_PROC(y.A), TO_LOCAL(y.A), size);
	} else if (TO_PROC(y.A) == MYPROC) {
	  /* remote<-local */
	  ti_bulk_write_weak(TO_PROC(x.A), TO_LOCAL(x.A), TO_LOCAL(y.A), size);
	} else {
	  /* remote<-remote */
	  unsigned char *buf;
	  buf = (unsigned char *) malloc(size);
	  ti_bulk_read(buf, TO_PROC(y.A), TO_LOCAL(y.A), size);
	  ti_sync();
	  ti_bulk_write_weak(TO_PROC(x.A), TO_LOCAL(x.A), buf, size);
	  free(buf);
	}
        ti_sync();

	return;
      }
    }
  }
#endif
#endif /* GLOBAL_ARRAY */

  {
    T tmp;
    int i, lo[N], hi[N], stride[N];
    ti_POINT p, q, s;

    /* Do an el-by-el copy. */
    p = RECTDOMAIN_MIN(R);
    q = RECTDOMAIN_MAX(R);
    s = RECTDOMAIN_STRIDE(R);
    for (i = 0; i < N; i++) {
      lo[i] = POINT_INDEX(p, i + 1);
      hi[i] = POINT_INDEX(q, i + 1);
      if ((stride[i] = POINT_INDEX(s, i + 1)) == 0) stride[i] = 1;
    }
    forall2(e, x, f, y, lo, hi, stride, deref(tmp, f); weak_assign(e, tmp));
    ti_sync();
  }
}

/*
 * x.set(w) is equivalent to:
 * foreach (p in x.domain()) x[p] = w;
 */
INLINE_MODIFIER void SET_METHOD(const ti_ARRAY x, T w)
{
  ti_RECTDOMAIN R = x.domain;
  ti_POINT p, q, s;
  jint lo[N], hi[N], stride[N], i;

  assertExists(x, "(in set method)");
  p = RECTDOMAIN_MIN(R);
  q = RECTDOMAIN_MAX(R);
  s = RECTDOMAIN_STRIDE(R);
  for (i = 0; i < N; i++) {
    lo[i] = POINT_INDEX(p, i + 1);
    hi[i] = POINT_INDEX(q, i + 1);
    if ((stride[i] = POINT_INDEX(s, i + 1)) == 0) stride[i] = 1;
  }
  
  forall(e, x, lo, hi, stride, assign(e, w));
}
#endif /* assignable */

INLINE_MODIFIER ti_ARRAY TRANSLATE_METHOD(ti_ARRAY x, ti_POINT p)
{
  ti_ARRAY n = x;
  int i;
  assertExists(x, "(in translate method)");
  n.domain = RECTDOMAIN_TRANSLATE(x.domain, p);
  p = RECTDOMAIN_MIN(n.domain);
  for (i = 0; i < N; i++)
    n.base[i] = POINT_INDEX(p, i + 1);

#if DEBUG_ARRAYS
#if 0
  fprintf(stderr, "p%d: translated array with domain <%d, %x> = ",
	  MYPROC, TO_PROC(n.domain), (int) TO_LOCAL(n.domain));
#endif
  PRINT_DOMAIN(n.domain, stderr);
  fprintf(stderr, "\nArray<%d>:\tbase\tstride\tsideFactors\n", N);
  for (i = 0; i < N; i++) {
    fprintf(stderr, " %d\t\t%d\t%d\t%d\n", i,
	    n.base[i], n.stride[i], n.sideFactors[i]);
  }
#endif

  return n;
}

#define PERMUTE_C_ARRAY(src, dest, p)						\
do {										\
  int i_;									\
										\
  for (i_ = 0; i_ < N; i_++) dest[POINT_INDEX(p, i_ + 1) - 1] = src[i_];	\
} while (0)

INLINE_MODIFIER ti_ARRAY PERMUTE_METHOD(ti_ARRAY x, ti_POINT p)
{
  ti_ARRAY n = x;
  assertExists(x, "(in permute method)");
  n.domain = RECTDOMAIN_PERMUTE(n.domain, p);
  PERMUTE_C_ARRAY(x.base, n.base, p);
  PERMUTE_C_ARRAY(x.stride, n.stride, p);
  PERMUTE_C_ARRAY(x.sideFactors, n.sideFactors, p);
  return n;
}

#undef PERMUTE_C_ARRAY

INLINE_MODIFIER ti_ARRAY INJECT_METHOD(ti_ARRAY x, ti_POINT p)
{
  ti_ARRAY n = x;
  int i;
  assertExists(x, "(in inject method)");
  n.domain = RECTDOMAIN_INJECT(x.domain, p);
  for (i = 0; i < N; i++) {
    int px = POINT_INDEX(p, i + 1);
    n.base[i] *= px;
    n.stride[i] *= px;
  }      

#if 0
#if 0
  fprintf(stderr, "p%d: injected array with domain <%d, %x> = ",
	  MYPROC, TO_PROC(n.domain), (int) TO_LOCAL(n.domain));
#endif
  PRINT_DOMAIN(n.domain, stderr);
  fprintf(stderr, "\nArray<%d>:\tbase\tstride\tsideFactors\n", N);
  for (i = 0; i < N; i++) {
    fprintf(stderr, " %d\t\t%d\t%d\t%d\n", i,
	    n.base[i], n.stride[i], n.sideFactors[i]);
  }
#endif

  return n;
}

INLINE_MODIFIER ti_ARRAY PROJECT_METHOD(ti_ARRAY x, ti_POINT p)
{
  ti_ARRAY n = x;
  int i;
  assertExists(x, "(in project method)");
  n.domain = RECTDOMAIN_PROJECT(x.domain, p);
  for (i = 0; i < N; i++) {
    int px = POINT_INDEX(p, i + 1);
    assert(("project() beyond limits", n.base[i] % px == 0));
    n.base[i] /= px;
    assert(("project() beyond limits", n.stride[i] % px == 0));
    n.stride[i] /= px;
  }      

#if 0
#if 0
  fprintf(stderr, "p%d: projected array with domain <%d, %x> = ",
	  MYPROC, TO_PROC(n.domain), (int) TO_LOCAL(n.domain));
#endif
  PRINT_DOMAIN(n.domain, stderr);
  fprintf(stderr, "\nArray<%d>:\tbase\tstride\tsideFactors\n", N);
  for (i = 0; i < N; i++) {
    fprintf(stderr, " %d\t\t%d\t%d\t%d\n", i,
	    n.base[i], n.stride[i], n.sideFactors[i]);
  }
#endif

  return n;
}

#define DEBUG_ISCONTIGUOUS 0
jboolean ISCONTIGUOUS_METHOD(const ti_ARRAY x) {
  assertExists(x, "(in isContiguous method)");
  {
  #if 0
    /* this test fails in the presence of negative strides */
    int numelements = RECTDOMAIN_NUMPOINTS(x.domain);
    int minoffset = ti_conv("isContiguous()", &x, RECTDOMAIN_MIN(x.domain));
    int maxoffset = ti_conv("isContiguous()", &x, RECTDOMAIN_MAX(x.domain));
    return (numelements == maxoffset - minoffset + 1);
  #else 
    /* this test handles negative strides */
    int i;
    ti_POINT pointwithminaddr, pointwithmaxaddr;
    ti_POINT dommin, dommax, domstride;

    dommin = RECTDOMAIN_MIN(x.domain);
    dommax = RECTDOMAIN_MAX(x.domain);
    domstride = RECTDOMAIN_STRIDE(x.domain);

    #if DEBUG_ISCONTIGUOUS
      fprintf(stderr, "\n");
      fprintf(stderr, "dommin=");
      PRINT_POINT(dommin, stderr);
      fprintf(stderr, "\ndommax=");
      PRINT_POINT(dommax, stderr);
      fprintf(stderr, "\ndomstride=");
      PRINT_POINT(domstride, stderr);
      fflush(stderr);
    #endif

    for (i = 0; i < N; i++) {
      int lo = POINT_INDEX(dommin, i + 1);
      int hi = POINT_INDEX(dommax, i + 1);
      #if DEBUG_ISCONTIGUOUS
        fprintf(stderr, "\nlo=%i",lo);
        fprintf(stderr, "\nhi=%i",hi);
      #endif
      if (POINT_INDEX(domstride, i + 1) >= 0) {
        pointwithminaddr = POINT_SET(pointwithminaddr, i, lo);
        pointwithmaxaddr = POINT_SET(pointwithmaxaddr, i, hi);
        }
      else {
        pointwithminaddr = POINT_SET(pointwithminaddr, i, hi);
        pointwithmaxaddr = POINT_SET(pointwithmaxaddr, i, lo);
        }
      }

    #if DEBUG_ISCONTIGUOUS
      fprintf(stderr, "\npointwithminaddr=");
      PRINT_POINT(pointwithminaddr, stderr);
      fprintf(stderr, "\npointwithmaxaddr=");
      PRINT_POINT(pointwithmaxaddr, stderr);
      fprintf(stderr, "\n");
      fflush(stderr);
    #endif

    {
      int numelements = RECTDOMAIN_NUMPOINTS(x.domain);
      int minoffset = ti_conv("isContiguous()", &x, pointwithminaddr);
      int maxoffset = ti_conv("isContiguous()", &x, pointwithmaxaddr);
      return (numelements == maxoffset - minoffset + 1);
      }
  #endif
  }
  }
#undef DEBUG_ISCONTIGUOUS

PTR_TO_T ti_get_array_data_ptr(const ti_ARRAY x) {
  /* returns a ptr to the element with the lowest address in the current array */
  /* (should be used to get a handle on the array data when performing bulk operations on contiguous arrays) */
  assertExists(x, "(in isContiguous method)");
  {
  ti_POINT pointwithminaddr;
  PTR_TO_T tiarray_firstelem;
  #if 0
    /* no negative strides */
    pointwithminaddr = RECTDOMAIN_MIN(x.domain);
  #else
    ti_POINT dommin, dommax, domstride;
    int i;

    dommin = RECTDOMAIN_MIN(x.domain);
    dommax = RECTDOMAIN_MAX(x.domain);
    domstride = RECTDOMAIN_STRIDE(x.domain);
    for (i = 0; i < N; i++) {
      if (POINT_INDEX(domstride, i + 1) >= 0) 
        pointwithminaddr = POINT_SET(pointwithminaddr, i, POINT_INDEX(dommin, i + 1));        
      else 
        pointwithminaddr = POINT_SET(pointwithminaddr, i, POINT_INDEX(dommax, i + 1));
      }
  #endif
  INDEX(tiarray_firstelem, x.A, ti_conv("ti_get_array_data_ptr", &x, pointwithminaddr));
  return tiarray_firstelem;
  }
  }

/* ------------------------------------------------------------------------------------ */
/* bulk I/O methods */
/* Note: the following methods allow reading/writing Ti arrays to RandomAccessFile objects and 
   DataInputStream/DataOutputStream objects. 

   Another design possibility is to replace the DIS/DOS methods with ones taking InputStream/OutputStream.
     I don't currently see an advantage either way - both solutions are equally expressive 
     and neither is clearly more efficient. I stuck with first solution because dispatch was easier.
   */

#ifndef __bulkio_typedefs
#define __bulkio_typedefs
  /* these three seem to always get generated */
  #include "layout!T16RandomAccessFile2io4java.h"
  #include "layout!T15DataInputStream2io4java.h"
  #include "layout!T16DataOutputStream2io4java.h"

  /* these three are more of a struggle */
  #ifdef _include_layout_PT16RandomAccessFile2io4java_h_
    #define GP_RAF PT16RandomAccessFile2io4java
  #else
    typedef GP_type( T16RandomAccessFile2io4java ) GP_RAF;
  #endif
  #ifdef _include_layout_PT15DataInputStream2io4java_h_
    #define GP_DIS PT15DataInputStream2io4java
  #else
    typedef GP_type( T15DataInputStream2io4java ) GP_DIS;
  #endif
  #ifdef _include_layout_PT16DataOutputStream2io4java_h_
    #define GP_DOS PT16DataOutputStream2io4java
  #else
    typedef GP_type( T16DataOutputStream2io4java ) GP_DOS;
  #endif
#endif

#if ELEMENTS_ARE_ATOMIC
  /* RandomAccessFile */
  #if assignable
    #define TIARRAY_BULKIO_READFUNC_HEADER \
	    void READFROMRAF_METHOD(const ti_ARRAY x, GP_RAF fileobj)
    #define TIARRAY_BULKIO_READFUNC_CALLIO(fileobj, array, offset, count) \
      m9readFullyPTAjbyteIImT16RandomAccessFile2io4java( (fileobj), (array), (offset), (count) )
  #endif

  #define TIARRAY_BULKIO_WRITEFUNC_HEADER \
	  void WRITETORAF_METHOD(const ti_ARRAY x, GP_RAF fileobj)
  #define TIARRAY_BULKIO_WRITEFUNC_CALLIO(fileobj, array, offset, count) \
    m5writePTAjbyteIImT16RandomAccessFile2io4java( (fileobj), (array), (offset), (count) )

  #include "ti_array_bulkio.c"

  #if assignable
    #undef TIARRAY_BULKIO_READFUNC_HEADER
    #undef TIARRAY_BULKIO_READFUNC_CALLIO
  #endif
  #undef TIARRAY_BULKIO_WRITEFUNC_HEADER
  #undef TIARRAY_BULKIO_WRITEFUNC_CALLIO

  /* DataInputStream/DataOutputStream */
  #if assignable
    #define TIARRAY_BULKIO_READFUNC_HEADER \
	    void READFROMDIS_METHOD(const ti_ARRAY x, GP_DIS fileobj)
    #define TIARRAY_BULKIO_READFUNC_CALLIO(fileobj, array, offset, count) \
      m9readFullyPTAjbyteIImT15DataInputStream2io4java( (fileobj), (array), (offset), (count) )
  #endif

  #define TIARRAY_BULKIO_WRITEFUNC_HEADER \
	  void WRITETODOS_METHOD(const ti_ARRAY x, GP_DOS fileobj)
  #define TIARRAY_BULKIO_WRITEFUNC_CALLIO(fileobj, array, offset, count) \
    m5writePTAjbyteIImT16DataOutputStream2io4java( (fileobj), (array), (offset), (count) )

  #include "ti_array_bulkio.c"

  #if assignable
    #undef TIARRAY_BULKIO_READFUNC_HEADER
    #undef TIARRAY_BULKIO_READFUNC_CALLIO
  #endif
  #undef TIARRAY_BULKIO_WRITEFUNC_HEADER
  #undef TIARRAY_BULKIO_WRITEFUNC_CALLIO
#endif
/* ------------------------------------------------------------------------------------ */


#undef POINT_PRINT
#undef N_LARGER_THAN_1
#undef LOCAL_PART
#undef PROC_PART
#undef deref
#undef INDEX
#undef assign
#undef weak_assign
#undef assignable
#undef array_ralloc
#undef array_malloc
