/* ti_array.h */
/* This file assumes T is a type and N is an arity. */

#include "runtime-options.h"
#include "assertExists.h"

/* Instructions on how to add a method are in ti_array.c */

#define N_LARGER_THAN_1 N_MINUS_1

#ifndef GLOBAL_ARRAY
#define GLOBAL_ARRAY 1
#endif

#ifndef DEBUG_ARRAYS
#define DEBUG_ARRAYS 0
#endif

#ifndef ASSERT_SAME_BYTES
#define ASSERT_SAME_BYTES
static void assertSameBytes(char *x, char *y, int len)
{
  while (len-- > 0) assert(*x++ == *y++);
}
#endif /* ASSERT_SAME_BYTES */

#ifndef DUMPABLE_ARRAYS
#define DUMPABLE_ARRAYS 0
#endif

#ifndef DEFAULT_DUMP_DEST
#define DEFAULT_DUMP_DEST "|/disks/orodruin/titanium/work/pike/bin/aview"
#endif

#ifdef MEMORY_SHARED
#define EXPLICITLY_STORE_CREATOR 1
#else
#define EXPLICITLY_STORE_CREATOR 0
#endif

#if GLOBAL_ARRAY
#define deref DEREF_GLOBAL
#define LOCAL_PART(p)	(TO_LOCAL(p))
#define PROC_PART(p)	(TO_PROC(p))
#define INDEX		INDEX_GLOBAL
#else
#define deref DEREF_LOCAL
#define LOCAL_PART(p) (p)
#define PROC_PART(p) MYPROC
#define INDEX		INDEX_LOCAL
#endif

#if GLOBAL_ARRAY && !GLOBAL_ELEMENTS
# define assignable 0
#else
# define assignable 1
#endif

#define PRINT_POINT(p, f) 					\
do {								\
  jint _i;							\
  fprintf(f, "[");						\
  for (_i = 1; _i <= N; _i++)					\
    fprintf(f, (_i > 1) ? ", %d" : "%d", POINT_INDEX(p, _i));	\
  fprintf(f, "]");						\
} while (0)

#define PRINT_DOMAIN(R, f)			\
do {						\
  jint _i;					\
  ti_POINT p = RECTDOMAIN_STRIDE(R);		\
  fputs("[", f);				\
  PRINT_POINT(RECTDOMAIN_MIN(R), f);		\
  fputs(" : ", f);				\
  PRINT_POINT(RECTDOMAIN_UPB(R), f);		\
  for (_i = 1; _i <= N; _i++)			\
    if (POINT_INDEX(p, _i) >= 2) break;		\
  if (_i <= N) {				\
    fputs(" : ", f);				\
    PRINT_POINT(p, f);				\
  }						\
  fputs("]", f);				\
} while (0)


/* Address of arr[p] is:
   desc.A + ((p[1] - desc.base[0]) / desc.stride[0]) * desc.sideFactors[0]
          + ((p[2] - desc.base[1]) / desc.stride[1]) * desc.sideFactors[1]
          + ...
*/

/* Functions defined in ti_array.cc */
extern void ti_dump(ti_ARRAY *px, char *dest, char *userArg);
extern ti_ARRAY ti_empty();
extern jboolean ti_isnull();
extern PTR_TO_T ti_get_array_data_ptr(const ti_ARRAY x);
extern ti_ARRAY ti_construct(const char *where, Region pregion, ti_RECTDOMAIN R);
#if N_LARGER_THAN_1
extern ti_ARRAY_SLICE SLICE_METHOD(const ti_ARRAY x, jint k, jint j);
#else
void SLICE_METHOD(const ti_ARRAY x, jint k, jint j);
#endif
#if assignable
extern void EXCHANGE_METHOD(const ti_ARRAY x, T w);
extern void COPY_METHOD(const ti_ARRAY x, const ti_ARRAY y);
void SET_METHOD(const ti_ARRAY x, T w);
#endif
ti_ARRAY TRANSLATE_METHOD(ti_ARRAY x, ti_POINT p);
ti_ARRAY PERMUTE_METHOD(ti_ARRAY x, ti_POINT p);
ti_ARRAY INJECT_METHOD(ti_ARRAY x, ti_POINT p);
ti_ARRAY PROJECT_METHOD(ti_ARRAY x, ti_POINT p);
jboolean ISCONTIGUOUS_METHOD(const ti_ARRAY x);

/* bulk I/O functions */
#ifndef __bulkio_typedefs
#define __bulkio_typedefs
  /* these three seem to always get generated */
  #include "layout!T16RandomAccessFile2io4java.h"
  #include "layout!T15DataInputStream2io4java.h"
  #include "layout!T16DataOutputStream2io4java.h"

  /* these three are more of a struggle */
  #ifdef _include_layout_PT16RandomAccessFile2io4java_h_
    #define GP_RAF PT16RandomAccessFile2io4java
  #else
    typedef GP_type( T16RandomAccessFile2io4java ) GP_RAF;
  #endif
  #ifdef _include_layout_PT15DataInputStream2io4java_h_
    #define GP_DIS PT15DataInputStream2io4java
  #else
    typedef GP_type( T15DataInputStream2io4java ) GP_DIS;
  #endif
  #ifdef _include_layout_PT16DataOutputStream2io4java_h_
    #define GP_DOS PT16DataOutputStream2io4java
  #else
    typedef GP_type( T16DataOutputStream2io4java ) GP_DOS;
  #endif
#endif

#if ELEMENTS_ARE_ATOMIC
  #if assignable
    void READFROMRAF_METHOD(const ti_ARRAY x, GP_RAF fileobj);
    void READFROMDIS_METHOD(const ti_ARRAY x, GP_DIS fileobj);
  #endif
  void WRITETORAF_METHOD(const ti_ARRAY x, GP_RAF fileobj);
  void WRITETODOS_METHOD(const ti_ARRAY x, GP_DOS fileobj);
#endif

#if BOUNDS_CHECKING
jint ti_conv(const char *where, const ti_ARRAY *x, ti_POINT p);
#else /* !BOUNDS_CHECKING */
static INLINE_MODIFIER jint ti_conv(const char *where, const ti_ARRAY *x, ti_POINT p)
{
  jint sum = 0, q, qm, s;
#if N_LARGER_THAN_1
  jint i;
#else
#define i 1
#endif

  assertExists(*x, where);
#if N_LARGER_THAN_1
  for (i = N; i >= 1; i--)
#endif
    {
      q = POINT_INDEX(p, i);
      qm = q - x->base[i-1];
      /* At this point, x->stride[i-1] should divide qm.  Of course, we only
	 know for sure if bounds checking is on. */
      sum += (qm / x->stride[i-1]) * x->sideFactors[i-1];
    }
#if !N_LARGER_THAN_1
#undef i  
#endif

#if 0
  fprintf(stderr, "conv(");
  PRINT_POINT(p, stderr);
  fprintf(stderr, ") = %d\n", sum);
#endif
  return sum;
}
#endif /* BOUNDS_CHECKING */

static INLINE_MODIFIER PTR_TO_T ti_ptr(const char *where,
				       const ti_ARRAY x, ti_POINT p) {
  PTR_TO_T result;
  INDEX(result, x.A, ti_conv(where, &x, p));
  return result;
}

static INLINE_MODIFIER T ti_read(const char *where,
				 const ti_ARRAY x, ti_POINT p) {
  PTR_TO_T ptr = ti_ptr(where, x, p);
  T result;
  deref(result, ptr);
  return result;
}

#if assignable
static INLINE_MODIFIER void ti_write(const char *where,
				     const ti_ARRAY x, ti_POINT p, T value) {
  PTR_TO_T ptr = ti_ptr(where, x, p);
  assign(ptr, value);
}
#endif

static INLINE_MODIFIER ti_RECTDOMAIN DOMAIN_METHOD(const ti_ARRAY x)
{
  assertExists(x, "(in domain method)");
  return x.domain;
}

static INLINE_MODIFIER ti_ARRAY RESTRICT_METHOD(const ti_ARRAY x, ti_RECTDOMAIN R)
{
  ti_ARRAY n = x;
  ti_POINT p;
  int i, max[N];

  assertExists(x, "(in restrict method)");
  n.domain = RECTDOMAIN_INTERSECTION(x.domain, R);
  return n;
}

static INLINE_MODIFIER jint CREATOR_METHOD(const ti_ARRAY x)
{
#if EXPLICITLY_STORE_CREATOR
  return x.creator;
#else
  return PROC_PART(x.A);
#endif
}

static INLINE_MODIFIER jboolean ISLOCAL_METHOD(const ti_ARRAY x)
{
#ifdef MEMORY_SHARED
  return 1;
#else
  return (CREATOR_METHOD(x) == MYPROC);
#endif
}

#undef POINT_PRINT
#undef N_LARGER_THAN_1
#undef LOCAL_PART
#undef PROC_PART
#undef INDEX
#undef deref
#undef assign
#undef weak_assign
#undef assignable
