//.TEMPLATE 2 2
// Uses fake templates.  Script replaces 2 with specified.

package ti.domains;

//
// =====
// Point
// =====
//
// This represents a point in Z^N space, where N is the "arity".
//

public immutable class tiPoint2 {

  // Constructors
  public tiPoint2() {
    val = new (tiDomains.region) int[2];
  }
  public tiPoint2(tiPoint2 p) {
    val = new (tiDomains.region) int[2];
    for (int i = 0; i < getArity(); i++) {
      val[i] = p.get(i);
    }
  }
  public tiPoint2(int value) {
    val = new (tiDomains.region) int[2];
    for (int i = 0; i < getArity(); i++) {
      val[i] = value;
    }
  }
  public tiPoint2(int dir, int k) {
    int direction = tiDomainUtils.direction_value(dir);

    val = new (tiDomains.region) int[2];
    for (int i = 0; i < getArity(); i++) {
      if (i == direction) {
	val[i] = k*(tiDomainUtils.direction_backward(dir) ? -1 : 1);
      } else {
	val[i] = 0;
      }
    }
  }

  // Information
  public int getArity() { return 2; }
  public int get(int index) {
    return val[index];
  }
  public int getPrevious(int index) {
    return val[index-1];
  }

  // Assignment
  public tiPoint2 setThis(int index, int value) {
    val[index] = value;
    return this;
  }
  public tiPoint2 setAllThis(int value) {
    for (int i = 0; i < getArity(); i++) {
      val[i] = value;
    }
    return this;
  }
  public tiPoint2 set(int index, int value) {
    tiPoint2 new_p = new tiPoint2(this);
    new_p.val[index] = value;
    return new_p;
  }
  public static tiPoint2 setDirection(int dir, int k) {
    return new tiPoint2(dir, k);
  }
  public static tiPoint2 setDirection(int dir) {
    return new tiPoint2(dir, 1);
  }
  public static tiPoint2 setAll(int val) {
    return new tiPoint2(val);
  }

  // Arithmetic relationships
  public tiPoint2 add(tiPoint2 p) {
    tiPoint2 new_p = new tiPoint2();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = val[i] + p.val[i];
    }
    return new_p;
  }
  public tiPoint2 sub(tiPoint2 p) {
    tiPoint2 new_p = new tiPoint2();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = val[i] - p.val[i];
    }
    return new_p;
  }
  public tiPoint2 mul(int value) {
    tiPoint2 new_p = new tiPoint2();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = val[i] * value;
    }
    return new_p;
  }
  public tiPoint2 mul(tiPoint2 p) {
    tiPoint2 new_p = new tiPoint2();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = val[i] * p.val[i];
    }
    return new_p;
  }
  public tiPoint2 div(int value) {
    return div(new tiPoint2(value));
  }
  public tiPoint2 div(tiPoint2 p) {
    tiPoint2 new_p = new tiPoint2();
    for (int i = 0; i < getArity(); i++) {
      if (p.val[i] == 0) {
	System.out.println("Runtime: (Error) division by 0");
	new_p.val[i] = 0;
      } else {
	// divide rounding toward zero from both directions
	if (val[i] < 0) {
	  new_p.val[i] = ((val[i] - (p.val[i]-1)) / p.val[i]);
	} else {
	  new_p.val[i] = val[i] / p.val[i];
	}
      }
    }
    return new_p;
  }
  public int dot(tiPoint2 p) {
    int result = 0;
    for (int i = 0; i < getArity(); i++) {
      result += (val[i] * p.val[i]);
    }
    return result;
  }

  // Boolean relationships
  public boolean isEqual(tiPoint2 p) {
    boolean result = true;
    for (int i = 0; (i < getArity()) && result; i++) {
      if (val[i] != p.val[i]) {
	result = false;
      }
    }
    return result;
  }
  public boolean isNotEqual(tiPoint2 p) {
    boolean result = true;
    for (int i = 0; (i < getArity()) && result; i++) {
      if (val[i] != p.val[i]) {
	return true;
      }
    }
    return false;
  }
  public boolean isLessThan(tiPoint2 p) {
    boolean result = true;
    for (int i = 0; (i < getArity()) && result; i++) {
      if (!(val[i] < p.val[i])) {
	result = false;
      }
    }
    return result;
  }
  public boolean isLessThanEqual(tiPoint2 p) {
    boolean result = true;
    for (int i = 0; (i < getArity()) && result; i++) {
      if (!(val[i] <= p.val[i])) {
	result = false;
      }
    }
    return result;
  }
  public boolean isGreaterThan(tiPoint2 p) {
    boolean result = true;
    for (int i = 0; (i < getArity()) && result; i++) {
      if (!(val[i] > p.val[i])) {
	result = false;
      }
    }
    return result;
  }
  public boolean isGreaterThanEqual(tiPoint2 p) {
    boolean result = true;
    for (int i = 0; (i < getArity()) && result; i++) {
      if (!(val[i] >= p.val[i])) {
	result = false;
      }
    }
    return result;
  }

  // Point-Point relationships
  public tiPoint2 getLowerBound(tiPoint2 p) {
    tiPoint2 new_p = new tiPoint2();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = (val[i] < p.val[i]) ? val[i] : p.val[i];
    }
    return new_p;
  }
  public tiPoint2 getUpperBound(tiPoint2 p) {
    tiPoint2 new_p = new tiPoint2();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = (val[i] > p.val[i]) ? val[i] : p.val[i];
    }
    return new_p;
  }
  public tiPoint2 getLcm(tiPoint2 p) {
    tiPoint2 new_p = new tiPoint2();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = tiDomainUtils.lcm(val[i], p.val[i]);
    }
    return new_p;
  }
  public tiPoint2 permute(tiPoint2 p) {
    tiPoint2 new_p = new tiPoint2();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = val[p.val[i]];
    }
    return new_p;
  }

  // State
  // private final int val[] = new int[2];
  private int val[];
}
