//.TEMPLATE 3 2
// Uses fake templates.  Script replaces 3 with specified.

package ti.domains;

//
// =====
// Point
// =====
//
// This represents a point in Z^N space, where N is the "arity".
//

public immutable class tiPoint3 {

  // Constructors
  public tiPoint3() {
    val = new (tiDomains.region) int[3];
  }
  public tiPoint3(tiPoint3 p) {
    val = new (tiDomains.region) int[3];
    for (int i = 0; i < getArity(); i++) {
      val[i] = p.get(i);
    }
  }
  public tiPoint3(int value) {
    val = new (tiDomains.region) int[3];
    for (int i = 0; i < getArity(); i++) {
      val[i] = value;
    }
  }
  public tiPoint3(int dir, int k) {
    int direction = tiDomainUtils.direction_value(dir);

    val = new (tiDomains.region) int[3];
    for (int i = 0; i < getArity(); i++) {
      if (i == direction) {
	val[i] = k*(tiDomainUtils.direction_backward(dir) ? -1 : 1);
      } else {
	val[i] = 0;
      }
    }
  }

  // Information
  public int getArity() { return 3; }
  public int get(int index) {
    return val[index];
  }
  public int getPrevious(int index) {
    return val[index-1];
  }

  // Assignment
  public tiPoint3 setThis(int index, int value) {
    val[index] = value;
    return this;
  }
  public tiPoint3 setAllThis(int value) {
    for (int i = 0; i < getArity(); i++) {
      val[i] = value;
    }
    return this;
  }
  public tiPoint3 set(int index, int value) {
    tiPoint3 new_p = new tiPoint3(this);
    new_p.val[index] = value;
    return new_p;
  }
  public static tiPoint3 setDirection(int dir, int k) {
    return new tiPoint3(dir, k);
  }
  public static tiPoint3 setDirection(int dir) {
    return new tiPoint3(dir, 1);
  }
  public static tiPoint3 setAll(int val) {
    return new tiPoint3(val);
  }

  // Arithmetic relationships
  public tiPoint3 add(tiPoint3 p) {
    tiPoint3 new_p = new tiPoint3();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = val[i] + p.val[i];
    }
    return new_p;
  }
  public tiPoint3 sub(tiPoint3 p) {
    tiPoint3 new_p = new tiPoint3();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = val[i] - p.val[i];
    }
    return new_p;
  }
  public tiPoint3 mul(int value) {
    tiPoint3 new_p = new tiPoint3();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = val[i] * value;
    }
    return new_p;
  }
  public tiPoint3 mul(tiPoint3 p) {
    tiPoint3 new_p = new tiPoint3();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = val[i] * p.val[i];
    }
    return new_p;
  }
  public tiPoint3 div(int value) {
    return div(new tiPoint3(value));
  }
  public tiPoint3 div(tiPoint3 p) {
    tiPoint3 new_p = new tiPoint3();
    for (int i = 0; i < getArity(); i++) {
      if (p.val[i] == 0) {
	System.out.println("Runtime: (Error) division by 0");
	new_p.val[i] = 0;
      } else {
	// divide rounding toward zero from both directions
	if (val[i] < 0) {
	  new_p.val[i] = ((val[i] - (p.val[i]-1)) / p.val[i]);
	} else {
	  new_p.val[i] = val[i] / p.val[i];
	}
      }
    }
    return new_p;
  }
  public int dot(tiPoint3 p) {
    int result = 0;
    for (int i = 0; i < getArity(); i++) {
      result += (val[i] * p.val[i]);
    }
    return result;
  }

  // Boolean relationships
  public boolean isEqual(tiPoint3 p) {
    boolean result = true;
    for (int i = 0; (i < getArity()) && result; i++) {
      if (val[i] != p.val[i]) {
	result = false;
      }
    }
    return result;
  }
  public boolean isNotEqual(tiPoint3 p) {
    boolean result = true;
    for (int i = 0; (i < getArity()) && result; i++) {
      if (val[i] != p.val[i]) {
	return true;
      }
    }
    return false;
  }
  public boolean isLessThan(tiPoint3 p) {
    boolean result = true;
    for (int i = 0; (i < getArity()) && result; i++) {
      if (!(val[i] < p.val[i])) {
	result = false;
      }
    }
    return result;
  }
  public boolean isLessThanEqual(tiPoint3 p) {
    boolean result = true;
    for (int i = 0; (i < getArity()) && result; i++) {
      if (!(val[i] <= p.val[i])) {
	result = false;
      }
    }
    return result;
  }
  public boolean isGreaterThan(tiPoint3 p) {
    boolean result = true;
    for (int i = 0; (i < getArity()) && result; i++) {
      if (!(val[i] > p.val[i])) {
	result = false;
      }
    }
    return result;
  }
  public boolean isGreaterThanEqual(tiPoint3 p) {
    boolean result = true;
    for (int i = 0; (i < getArity()) && result; i++) {
      if (!(val[i] >= p.val[i])) {
	result = false;
      }
    }
    return result;
  }

  // Point-Point relationships
  public tiPoint3 getLowerBound(tiPoint3 p) {
    tiPoint3 new_p = new tiPoint3();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = (val[i] < p.val[i]) ? val[i] : p.val[i];
    }
    return new_p;
  }
  public tiPoint3 getUpperBound(tiPoint3 p) {
    tiPoint3 new_p = new tiPoint3();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = (val[i] > p.val[i]) ? val[i] : p.val[i];
    }
    return new_p;
  }
  public tiPoint3 getLcm(tiPoint3 p) {
    tiPoint3 new_p = new tiPoint3();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = tiDomainUtils.lcm(val[i], p.val[i]);
    }
    return new_p;
  }
  public tiPoint3 permute(tiPoint3 p) {
    tiPoint3 new_p = new tiPoint3();
    for (int i = 0; i < getArity(); i++) {
      new_p.val[i] = val[p.val[i]];
    }
    return new_p;
  }

  // State
  // private final int val[] = new int[3];
  private int val[];
}
