// WARNING: This file is automatically generated.
// Uses faked templates.  Replace 3 with specified.

package ti.domains;

//
// ===========
// Rect Domain
// ===========
//
// This represents a domain that looks like an N-dim rectangle,
// with sides parallel to the dimensional axes.  This is optimized
// for a fast "foreach" at the expense of set operations.
//

public immutable class tiRectDomain3 {

  // Constructors
  public tiRectDomain3() {
    Point<3> zero_point = Point<3>.all(0);
    p0 = zero_point;
    p1 = zero_point;
    stride = zero_point;
    loopStride = Point<3>.all(1);
  }
  public tiRectDomain3(Point<3> set_p0,
			     Point<3> set_p1,
			     Point<3> set_stride) {
    Point<3> set_loopStride;

    // check stride is positive
    for (int i = 0; i < set_stride.arity(); i++) {
      if (set_stride.get(i) < 0) {
	// !!! error - stride may not contain negative Point<N> value
	System.out.println("Runtime: (Error) stride may not contain a negative Point<N> value.");
	break;
      }
    }

    // normalize
    if (set_p0.isLessThan(set_p1)) {
      // p0
      // nothing necessary

      // p1
      for (int i = 0; i < set_p1.arity(); i++) {
	if (set_stride.get(i) == 0) {
	  set_p1 = set_p1.setThis(i, set_p0.get(i) + 1);
	} else {
	  if (set_p1.get(i) > set_p0.get(i)) {
	    set_p1 = set_p1.setThis(i,
				    set_p0.get(i) +
				      tiDomainUtils.
				      align_down((set_p1.get(i) -
						  set_p0.get(i)) - 1,
						 set_stride.get(i)) + 1);
	  } else {
	    set_p1 = set_p1.setThis(i, set_p0.get(i));
	  }
	}
      }

      // stride
      for (int i = 0; i < set_stride.arity(); i++) {
	if (set_stride.get(i) >= (set_p1.get(i) - set_p0.get(i))) {
	  set_stride = set_stride.setThis(i, 0);
	}
      }

      // loop stride
      set_loopStride = set_stride;
      for (int i = 0; i < getArity(); i++) {
	if (set_loopStride.get(i) == 0) {
	  set_loopStride = set_loopStride.setThis(i, 1);
	}
      }
    } else {
      // empty set
      Point<3> zero_point = Point<3>.all(0);

      set_p0 = zero_point;
      set_p1 = zero_point;
      set_stride = zero_point;
      set_loopStride = Point<3>.all(1);
    }

    p0 = set_p0;
    p1 = set_p1;
    stride = set_stride;
    loopStride = set_loopStride;
  }
  public tiRectDomain3(tiRectDomain3 set_rd) {
    p0 = set_rd.p0;
    p1 = set_rd.p1;
    stride = set_rd.stride;
    loopStride = set_rd.loopStride;
  }
  public tiRectDomain3(tiDomain3 d) {
    tiRectDomainList3 rdl = d.demote().getRectangles();

    if (!rdl.hasSingularity()) {
      System.out.println("Runtime: (Error) not a RectDomain in cast.");
      Point<3> zero_point = Point<3>.all(0);
      p0 = zero_point;
      p1 = zero_point;
      stride = zero_point;
      loopStride = Point<3>.all(1);
    } else {
      tiRectDomain3 rd = rdl.getSingularity();

      p0 = rd.p0;
      p1 = rd.p1;
      stride = rd.stride;
      loopStride = rd.loopStride;
    }
  }
  public static tiRectDomain3 createWithMaxPoint(Point<3> set_p0,
						       Point<3> set_p1,
						       Point<3> set_stride) {
    return new tiRectDomain3(set_p0,
				   set_p1.add(Point<3>.all(1)),
				   set_stride);
  }
  public static tiRectDomain3 createWithUpbPoint(Point<3> set_p0,
						       Point<3> set_p1,
						       Point<3> set_stride) {
    return new tiRectDomain3(set_p0, set_p1, set_stride);
  }
  public static tiRectDomain3 createTranspose(tiRectDomain3 rd,
						    Point<3> tr) {
    return new tiRectDomain3(rd.p0.permute(tr),
				   rd.p1.permute(tr),
				   rd.stride.permute(tr));
  }

  // Set Relations
  public tiDomain3 add(tiRectDomain3 rd) {
    // this + d

    // do union
    if (isSubset(rd)) {
      return rd.promote();
    } else if (rd.isSubset(this)) {
      return this.promote();
    } else {
      // Can assume this and d are not subsets of each other.  Note that this
      // assumption implies that neither are null sets.

      // information to create result domain
      int expand_left = 0;
      int expand_right = 0;
      int set_stride = -1;
      boolean has_change_dim = false;
      int change_dim;

      // "non encompassing" is where d does not encompass this
      // a "hand" of 0 is left, 1 is middle, 2 is right
      int num_nonencomp_dim = 0;
      int last_nonencomp_dim;
      int last_nonencomp_dim_hand;
      // non overlap is where d does not overlap this
      // a "hand" of 0 is left, 2 is right
      int num_nonoverlap_dim = 0;
      int last_nonoverlap_dim;
      int last_nonoverlap_dim_hand;
      // a dimension where d does not equal this
      int num_unequal_dim = 0;
      int last_unequal_dim;

      // collect data
      for (int i = 0; i < getArity(); i++) {
	if (p1.get(i) <= rd.p0.get(i)) {
	  // cases:
	  // +- this -+
	  //            +- d -+
	  num_nonoverlap_dim++;
	  last_nonoverlap_dim = i;
	  last_nonoverlap_dim_hand = 2;
	  num_unequal_dim++;
	  last_unequal_dim = i;
	} else if (p0.get(i) >= rd.p1.get(i)) {
	  // cases:
	  //         +- this -+
	  // +- d -+
	  num_nonoverlap_dim++;
	  last_nonoverlap_dim = i;
	  last_nonoverlap_dim_hand = 0;
	  num_unequal_dim++;
	  last_unequal_dim = i;
	} else if ((p0.get(i) >= rd.p0.get(i)) &&
		   (p1.get(i) <= rd.p1.get(i))) {
	  // cases:
	  //   +- this -+
	  // +---- d -----+
	  // - or -
	  // +- this -+
	  // +---- d ----+
	  // - or -
	  //    +- this -+
	  // +---- d ----+
	  // - or -
	  // +- this -+
	  // +-- d ---+
	  if (!((p0.get(i) == rd.p0.get(i)) &&
		(p1.get(i) == rd.p1.get(i)))) {
	    num_unequal_dim++;
	    last_unequal_dim = i;
	  }
	} else {
	  num_unequal_dim++;
	  last_unequal_dim = i;
	  num_nonencomp_dim++;
	  last_nonencomp_dim = i;
	  if ((p0.get(i) < rd.p0.get(i)) && (p1.get(i) > rd.p1.get(i))) {
	    // cases:
	    // +-- this --+
	    //   +- d -+
	    last_nonencomp_dim_hand = 1;
	  } else if ((p0.get(i) < rd.p0.get(i)) &&
		     (p1.get(i) <= rd.p1.get(i))) {
	    // cases:
	    // +- this -+
	    //        +- d -+
	    // - or -
	    // +- this -+
	    //    +- d -+
	    last_nonencomp_dim_hand = 2;
	  } else {
	    // cases:
	    //    +- this -+
	    // +- d -+
	    // - or -
	    // +- this -+
	    // +- d -+
	    last_nonencomp_dim_hand = 0;
	  }
	}
      }

      if ((num_nonoverlap_dim == 1) && (num_unequal_dim == 1)) {
	if (stride.get(last_nonoverlap_dim) == 0) {
	  if (rd.stride.get(last_nonoverlap_dim) == 0) {
	    if (rd.p0.get(last_nonoverlap_dim) < p0.get(last_nonoverlap_dim)) {
	      // initiate stride left (stride regeneration)
	      expand_left = (p0.get(last_nonoverlap_dim) -
			     rd.p0.get(last_nonoverlap_dim));
	      set_stride = expand_left;
	    } else {
	      // initiate stride right (stride regeneration)
	      expand_right = (rd.p0.get(last_nonoverlap_dim) -
			      p0.get(last_nonoverlap_dim));
	      set_stride = expand_right;
	    }
	    has_change_dim = true;
	    change_dim = last_nonoverlap_dim;
	  } else {
	    if (p0.get(last_nonoverlap_dim) ==
		(rd.p0.get(last_nonoverlap_dim) -
		 rd.stride.get(last_nonoverlap_dim))) {
	      // right append to stride (stride regeneration)
	      expand_right = (rd.p1.get(last_nonoverlap_dim) -
			      1 -
			      rd.p0.get(last_nonoverlap_dim) +
			      rd.stride.get(last_nonoverlap_dim));
	      set_stride = rd.stride.get(last_nonoverlap_dim);
	      has_change_dim = true;
	      change_dim = last_nonoverlap_dim;
	    } else if (p0.get(last_nonoverlap_dim) ==
		       (rd.p1.get(last_nonoverlap_dim) +
			rd.stride.get(last_nonoverlap_dim) - 1)) {
	      // left append to stride (stride regeneration)
	      expand_left = (rd.p1.get(last_nonoverlap_dim) -
			     1 -
			     rd.p0.get(last_nonoverlap_dim) +
			     rd.stride.get(last_nonoverlap_dim));
	      set_stride = rd.stride.get(last_nonoverlap_dim);
	      has_change_dim = true;
	      change_dim = last_nonoverlap_dim;
	    }
	  }
	} else {
	  if ((rd.stride.get(last_nonoverlap_dim) ==
	       stride.get(last_nonoverlap_dim)) ||
	      (rd.stride.get(last_nonoverlap_dim) == 0)) {
	    if (last_nonoverlap_dim_hand == 0) {
	      if ((rd.p1.get(last_nonoverlap_dim) - 1) ==
		  (p0.get(last_nonoverlap_dim) -
		   stride.get(last_nonoverlap_dim))) {
		// abut left
		expand_left = (p0.get(last_nonoverlap_dim) -
			       rd.p0.get(last_nonoverlap_dim));
		has_change_dim = true;
		change_dim = last_nonoverlap_dim;
	      }
	    } else if (last_nonoverlap_dim_hand == 2) {
	      if (rd.p0.get(last_nonoverlap_dim) ==
		  (p1.get(last_nonoverlap_dim) -
		   1 +
		   stride.get(last_nonoverlap_dim))) {
		// abut right
		expand_right = (rd.p1.get(last_nonoverlap_dim) -
				p1.get(last_nonoverlap_dim));
		has_change_dim = true;
		change_dim = last_nonoverlap_dim;
	      }
	    }
	  }
	}
      }
      if ((!has_change_dim) &&
	  (num_nonencomp_dim == 1) &&
	  (num_unequal_dim == 1)) {
	if ((stride.get(last_nonencomp_dim) != 0) &&
	    (rd.stride.get(last_nonencomp_dim) ==
	     stride.get(last_nonencomp_dim))) {
	  if (((rd.p0.get(last_nonencomp_dim) -
		p0.get(last_nonencomp_dim)) %
	       stride.get(last_nonencomp_dim)) == 0) {
	    if (last_nonencomp_dim_hand == 0) {
	      // expand left equal stride
	      expand_left = (p0.get(last_nonencomp_dim) -
			     rd.p0.get(last_nonencomp_dim));
	      has_change_dim = true;
	      change_dim = last_nonencomp_dim;
	    } else if (last_nonencomp_dim_hand == 2) {
	      // expand right equal stride
	      expand_right = (rd.p1.get(last_nonencomp_dim) -
			      p1.get(last_nonencomp_dim));
	      has_change_dim = true;
	      change_dim = last_nonencomp_dim;
	    }
	  }
	}
      }
      if ((!has_change_dim) &&
	  (num_nonencomp_dim == 1) &&
	  (num_unequal_dim == 1)) {
	if (stride.get(last_nonencomp_dim) > 0) {
	  if ((rd.stride.get(last_nonencomp_dim) == 0)
	      ? false
	      : (((stride.get(last_nonencomp_dim) %
		   rd.stride.get(last_nonencomp_dim)) == 0) &&
		 (rd.stride.get(last_nonencomp_dim) <
		  stride.get(last_nonencomp_dim)))) {
	    if (last_nonencomp_dim_hand == 0) {
	      if ((rd.p0.get(last_nonencomp_dim) <=
		   p0.get(last_nonencomp_dim)) &&
		  (rd.p1.get(last_nonencomp_dim) ==
		   (p1.get(last_nonencomp_dim) -
		    rd.stride.get(last_nonencomp_dim)))) {
		// overlapped left interpolated points (stride reduction)
		expand_left = (p0.get(last_nonencomp_dim) -
			       rd.p0.get(last_nonencomp_dim));
		set_stride = rd.stride.get(last_nonencomp_dim);
		has_change_dim = true;
		change_dim = last_nonencomp_dim;
	      }
	    } else if (last_nonencomp_dim_hand == 2) {
	      if ((rd.p1.get(last_nonencomp_dim) >=
		   p1.get(last_nonencomp_dim)) &&
		  (rd.p0.get(last_nonencomp_dim) ==
		   (p0.get(last_nonencomp_dim) +
		    rd.stride.get(last_nonencomp_dim)))) {
		// overlapped right interpolated points (stride reduction)
		expand_right = (rd.p1.get(last_nonencomp_dim) -
				p1.get(last_nonencomp_dim));
		set_stride = rd.stride.get(last_nonencomp_dim);
		has_change_dim = true;
		change_dim = last_nonencomp_dim;
	      }
	    } else {
	      if ((rd.p0.get(last_nonencomp_dim) ==
		   (p0.get(last_nonencomp_dim) +
		    rd.stride.get(last_nonencomp_dim))) &&
		  (rd.p1.get(last_nonencomp_dim) ==
		   (p1.get(last_nonencomp_dim) -
		    rd.stride.get(last_nonencomp_dim)))) {
		// overlapped middle interpolated points (stride reduction)
		set_stride = rd.stride.get(last_nonencomp_dim);
		has_change_dim = true;
		change_dim = last_nonencomp_dim;
	      }
	    }
	  }
	}
      }
      if ((!has_change_dim) &&
	  (num_unequal_dim == 1)) {
	if (stride.get(last_unequal_dim) > 0) {
	  if ((stride.get(last_unequal_dim) % 2) == 0) {
	    int check_stride = stride.get(last_unequal_dim) / 2;
	    if ((rd.stride.get(last_unequal_dim) ==
		stride.get(last_unequal_dim)) ||
		(rd.stride.get(last_unequal_dim) == 0)) {
	      if ((rd.p0.get(last_unequal_dim) ==
		   (p0.get(last_unequal_dim) -
		    check_stride)) &&
		  (rd.p1.get(last_unequal_dim) ==
		   (p1.get(last_unequal_dim) -
		    check_stride))) {
		// non-overlapped left interpolated points (stride reduction)
		expand_left = check_stride;
		set_stride = check_stride;
		has_change_dim = true;
		change_dim = last_unequal_dim;
	      } else if ((rd.p0.get(last_unequal_dim) ==
			  (p0.get(last_unequal_dim) +
			   check_stride)) &&
			 (rd.p1.get(last_unequal_dim) ==
			  (p1.get(last_unequal_dim) +
			   check_stride))) {
		// non-overlapped right interpolated points (stride reduction)
		expand_right = check_stride;
		set_stride = check_stride;
		has_change_dim = true;
		change_dim = last_unequal_dim;
	      } else if ((rd.p0.get(last_unequal_dim) ==
			  (p0.get(last_unequal_dim) -
			   check_stride)) &&
			 (rd.p1.get(last_unequal_dim) ==
			  (p1.get(last_unequal_dim) +
			   check_stride))) {
		// non-overlapped left/right interpolated points (stride reduction)
		expand_left = check_stride;
		expand_right = check_stride;
		set_stride = check_stride;
		has_change_dim = true;
		change_dim = last_unequal_dim;
	      } else if ((rd.p0.get(last_unequal_dim) ==
			  (p0.get(last_unequal_dim) +
			   check_stride)) &&
			 (rd.p1.get(last_unequal_dim) ==
			  (p1.get(last_unequal_dim) -
			   check_stride))) {
		// non-overlapped middle interpolated points (stride reduction)
		set_stride = check_stride;
		has_change_dim = true;
		change_dim = last_unequal_dim;
	      }
	    }
	  }
	}
      }
      if ((!has_change_dim) &&
	  (num_unequal_dim == 1)) {
	if ((stride.get(last_unequal_dim) != 0) &&
	    (rd.stride.get(last_unequal_dim) != 0)) {
	  if (((rd.stride.get(last_unequal_dim) %
		stride.get(last_unequal_dim)) == 0) &&
	      (rd.stride.get(last_unequal_dim) >
		stride.get(last_unequal_dim))) {
	    if ((rd.p0.get(last_unequal_dim) ==
		 (p0.get(last_unequal_dim) - stride.get(last_unequal_dim))) &&
		(rd.p1.get(last_unequal_dim) <= p1.get(last_unequal_dim))) {
	      // expand left greater stride
	      expand_left = stride.get(last_unequal_dim);
	      has_change_dim = true;
	      change_dim = last_unequal_dim;
	    } else if ((rd.p0.get(last_unequal_dim) >=
			p0.get(last_unequal_dim)) &&
		       (rd.p1.get(last_unequal_dim) ==
			(p1.get(last_unequal_dim) +
			 stride.get(last_unequal_dim)))) {
	      // expand right greater stride
	      expand_right = stride.get(last_unequal_dim);
	      has_change_dim = true;
	      change_dim = last_unequal_dim;
	    } else if ((rd.p0.get(last_unequal_dim) ==
			(p0.get(last_unequal_dim) -
			 stride.get(last_unequal_dim))) &&
		       (rd.p1.get(last_unequal_dim) ==
			(p1.get(last_unequal_dim) +
			 stride.get(last_unequal_dim)))) {
	      // expand left/right greater stride
	      expand_left = stride.get(last_unequal_dim);
	      expand_right = stride.get(last_unequal_dim);
	      has_change_dim = true;
	      change_dim = last_unequal_dim;
	    }
	  }
	}
      }
      if ((!has_change_dim) &&
	  (num_unequal_dim == 1)) {
	if ((stride.get(last_unequal_dim) == 0) &&
	    ((rd.stride.get(last_unequal_dim) % 2) == 0)) {
	  int check_stride = rd.stride.get(last_unequal_dim) / 2;
	  if ((rd.p0.get(last_unequal_dim) ==
	       (p0.get(last_unequal_dim) - check_stride)) &&
	      (rd.p1.get(last_unequal_dim) ==
	       (p1.get(last_unequal_dim) + check_stride))) {
	      // initiate stride left/right (stride regeneration)
	      expand_left = check_stride;
	      expand_right = check_stride;
	      set_stride = check_stride;
	      has_change_dim = true;
	      change_dim = last_unequal_dim;
	    }
	  
	}
      }

      // execute changes
      if (has_change_dim) {
	Point<3> new_p0 = p0;
	Point<3> new_p1 = p1;
	Point<3> new_stride = stride;

	if (set_stride >= 0) {
	  new_stride = new_stride.set(change_dim, set_stride);
	}
	if (expand_left > 0) {
	  new_p0 =
	    new_p0.set(change_dim,
		       (new_p0.get(change_dim) -
			tiDomainUtils.
		          align_up(expand_left,
				   (new_stride.get(change_dim) != 0)
				     ? new_stride.get(change_dim)
				     : 1)));
	}
	if (expand_right > 0) {
	  new_p1 = new_p1.set(change_dim,
			      new_p1.get(change_dim) + expand_right);
	}
	return (new tiRectDomain3(new_p0, new_p1, new_stride)).promote();
      } else {
	return promote().add(rd);
      }
    }
  }
  public tiDomain3 difference(tiRectDomain3 rd) {
    // this - d

    // do difference
    // information to create result domain
    int shrink_left = 0;
    int shrink_right = 0;
    int mult_stride = 1;
    boolean has_change_dim = false;
    int change_dim;

    tiRectDomain3 di = multiply(rd);

    if (di.isNull()) {
      // sets disjoint
      return this.promote();
    } else if (di.isEqual(this)) {
      // null set
      return (new tiRectDomain3()).promote();
    }

    // "non encompassing" is where di does not encompass this
    // a "hand" of 0 is left, 1 is middle, 2 is right
    int num_nonencomp_dim = 0;
    int last_nonencomp_dim;
    int last_nonencomp_dim_hand;
    // when a dimension has the same limits, set if stride of di is 2x of this
    int num_equal_dim_stride_di_2x_this = 0;
    int last_equal_dim_stride_di_2x_this = 0;

    // collect data
    for (int i = 0; i < getArity(); i++) {
      if ((p1.get(i) <= di.p0.get(i)) || (p0.get(i) >= di.p1.get(i))) {
	// cases:
	// +- this -+
	//            +- di -+
	// - or -
	//          +- this -+
	// +- di -+
      } else if ((p0.get(i) >= di.p0.get(i)) && (p1.get(i) <= di.p1.get(i))) {
	// cases:
	//   +- this -+
	// +---- di ----+
	// - or -
	// +- this -+
	// +---- di ----+
	// - or -
	//     +- this -+
	// +---- di ----+
	// - or -
	// +- this -+
	// +-- di --+
	if ((p0.get(i) == di.p0.get(i)) && (p1.get(i) == di.p1.get(i))) {
	  if (stride.get(i) != 0) {
	    if (di.stride.get(i) == (2*(stride.get(i)))) {
	      num_equal_dim_stride_di_2x_this++;
	      last_equal_dim_stride_di_2x_this = i;
	    }
	  }
	}
      } else {
	num_nonencomp_dim++;
	last_nonencomp_dim = i;
	if ((p0.get(i) < di.p0.get(i)) && (p1.get(i) > di.p1.get(i))) {
	  // cases:
	  // +-- this --+
	  //   +- di -+
	  last_nonencomp_dim_hand = 1;
	} else if ((p0.get(i) < di.p0.get(i)) && (p1.get(i) <= di.p1.get(i))) {
	  // cases:
	  // +- this -+
	  //        +- di -+
	  // - or -
	  // +- this -+
	  //   +- di -+
	  last_nonencomp_dim_hand = 2;
	} else {
	  // cases:
	  //    +- this -+
	  // +- di -+
	  // - or -
	  // +- this -+
	  // +- di -+
	  last_nonencomp_dim_hand = 0;
	}
      }
    }

    // interpret collected data
    if (num_nonencomp_dim == 1) {
      if ((last_nonencomp_dim_hand == 0) &&
	  ((stride.get(last_nonencomp_dim) ==
	    di.stride.get(last_nonencomp_dim)) ||
	   di.stride.get(last_nonencomp_dim) == 0)) {
	// slice off left side
	shrink_left = (di.p1.get(last_nonencomp_dim) -
		       p0.get(last_nonencomp_dim));
	has_change_dim = true;
	change_dim = last_nonencomp_dim;
      } else if ((last_nonencomp_dim_hand == 2) &&
		 ((stride.get(last_nonencomp_dim) ==
		   di.stride.get(last_nonencomp_dim)) ||
		  di.stride.get(last_nonencomp_dim) == 0)) {
	// slice off right side
	shrink_right = (p1.get(last_nonencomp_dim) -
			di.p0.get(last_nonencomp_dim));
	has_change_dim = true;
	change_dim = last_nonencomp_dim;
      } else if ((last_nonencomp_dim_hand == 1) &&
		 ((stride.get(last_nonencomp_dim) == 0)
		    ? false
		    : (new tiRectDomain3
		        (p0.set(last_nonencomp_dim,
				(p0.get(last_nonencomp_dim) +
				 stride.get(last_nonencomp_dim))),
			 p1.set(last_nonencomp_dim,
				(p1.get(last_nonencomp_dim) -
				 stride.get(last_nonencomp_dim))),
			 stride)).isEqual(di))) {
	// middle elimination (stride magnification)
	mult_stride = ((p1.get(last_nonencomp_dim) -
			p0.get(last_nonencomp_dim) - 1) /
		       stride.get(last_nonencomp_dim));
	has_change_dim = true;
	change_dim = last_nonencomp_dim;
      }
    }
    if (!has_change_dim) {
      if (num_nonencomp_dim <= 1) {
	boolean has_check_dim = false;
	if (num_nonencomp_dim == 1) {
	  change_dim = last_nonencomp_dim;
	  has_check_dim = true;
	} else {
	  if (num_equal_dim_stride_di_2x_this == 1) {
	    change_dim = last_equal_dim_stride_di_2x_this;
	    has_check_dim = true;
	  }
	}
	if (has_check_dim) {
	  if ((stride.get(change_dim) != 0) &&
	      (di.stride.get(change_dim) != 0)) {
	    if ((stride.get(change_dim)*2) == di.stride.get(change_dim)) {
	      if ((di.p1.get(change_dim) == p1.get(change_dim)) ||
		  (di.p1.get(change_dim) ==
		   (p1.get(change_dim) - stride.get(change_dim)))) {
		if (p0.get(change_dim) == di.p0.get(change_dim)) {
		  // shifted interpolated points (stride magnification)
		  shrink_left = stride.get(change_dim);
		  mult_stride = 2;
		  has_change_dim = true;
		} else if ((p0.get(change_dim) + stride.get(change_dim)) ==
			   di.p0.get(change_dim)) {
		  // interpolated points (stride magnification)
		  mult_stride = 2;
		  has_change_dim = true;
		}
	      }
	    }
	  }
	}
      }
    }

    // execute changes
    if (has_change_dim) {
      Point<3> new_p0 = p0;
      Point<3> new_p1 = p1;
      Point<3> new_stride = stride;

      if (shrink_left > 0) {
	new_p0 =
	  new_p0.set(change_dim,
		     (new_p0.get(change_dim) +
		      tiDomainUtils.
		        align_up(shrink_left,
				 (stride.get(change_dim) != 0)
				   ? stride.get(change_dim)
				   : 1)));
      }
      if (shrink_right > 0) {
	new_p1 = new_p1.set(change_dim,
			    new_p1.get(change_dim) - shrink_right);
      }
      if (mult_stride > 1) {
	new_stride = new_stride.set(change_dim,
				    (new_stride.get(change_dim) *
				     mult_stride));
      }
      return (new tiRectDomain3(new_p0, new_p1, new_stride)).promote();
    } else {
      return promote().difference(rd);
    }
  }
  public tiRectDomain3 multiply(tiRectDomain3 rd) {
    // this * d

    // Intersection of a single rectangle with a single rectangle

    Point<3> new_p0;
    Point<3> new_p1;
    Point<3> new_stride;

    if (p0.isEqual(rd.p0)) {
      // p0 equal
      new_p0 = p0;
      new_p1 = p1.getLowerBound(rd.p1);
      new_stride = stride.getLcm(rd.stride);
    } else {
      // p0 not equal
      boolean detect_null_set = false;

      for (int i = 0; (i < new_p0.arity()) && (!detect_null_set); i++) {
	int start_p0 = p0.get(i);
	int start_d_p0 = rd.p0.get(i);
	int max_start = tiDomainUtils.max(start_p0, start_d_p0);

	if (rd.stride.get(i) != 0) {
	  start_d_p0 = (start_d_p0 +
			tiDomainUtils.align_up((max_start - start_d_p0),
					       rd.stride.get(i)));
	}
	if (stride.get(i) != 0) {
	  start_p0 = (start_p0 +
		      tiDomainUtils.
		        align_up((tiDomainUtils.max(max_start, start_d_p0) -
				  start_p0),
				 stride.get(i)));
	}

	if ((stride.get(i) == 0) || (rd.stride.get(i) == 0)) {
	  if (start_p0 != start_d_p0) {
	    detect_null_set = true;
	  } else {
	    new_p0 = new_p0.setThis(i, start_p0);
	  }
	} else {
	  boolean found = false;
	  for (int k = 0; (k < stride.get(i)) && (!found); k++) {
	    if (((rd.stride.get(i) * k) % stride.get(i)) ==
		((start_p0 - start_d_p0) % stride.get(i))) {
	      found = true;
	      new_p0 = new_p0.setThis(i, start_d_p0 + k * rd.stride.get(i));
	    }
	  }
	  if (!found) {
	    detect_null_set = true;
	  }
	}
      }

      if (detect_null_set) {
	// null set
	new_p0 = Point<3>.all(0);
	new_p1 = new_p0;
	new_stride = new_p0;
      } else {
	new_p1 = p1.getLowerBound(rd.p1);
	new_stride = stride.getLcm(rd.stride);
      }
    }

    return new tiRectDomain3(new_p0, new_p1, new_stride);
  }

  public boolean isNull() {
    // this is null
    Point<3> zero_point = Point<3>.all(0);
    return (p0.isEqual(zero_point) &&
	    p1.isEqual(zero_point) &&
	    stride.isEqual(zero_point));
  }
  public boolean isEqual(tiRectDomain3 rd) {
    // this == d

    return (p0.isEqual(rd.p0) &&
	    p1.isEqual(rd.p1) &&
	    stride.isEqual(rd.stride));
  }
  public boolean isNotEqual(tiRectDomain3 rd) {
    return !isEqual(rd);
  }
  public boolean isSubset(tiRectDomain3 rd) {
    // this <= d

    boolean stride_mult = true;

    for (int i = 0; (i < stride.arity()) && stride_mult; i++) {
      if (rd.stride.get(i) == 0) {
	if (stride.get(i) != 0) {
	  stride_mult = false;
	}
      } else if ((stride.get(i) % rd.stride.get(i)) != 0) {
	stride_mult = false;
      }
    }

    return ((stride_mult &&
	     p0.isGreaterThanEqual(rd.p0) &&
	     p1.isLessThanEqual(rd.p1) &&
	     rd.doesContain(p0)) ||
	    isNull());
  }
  public boolean isStrictSubset(tiRectDomain3 rd) {
    // this < d

    return isSubset(rd) && !isEqual(rd);
  }
  public boolean isSuperset(tiRectDomain3 rd) {
    return rd.isSubset(this);
  }
  public boolean isStrictSuperset(tiRectDomain3 rd) {
    return rd.isStrictSubset(this);
  }

  // Domain relations
  public tiDomain3 add(tiDomain3 d) {
    return promote().add(d);
  }
  public tiDomain3 difference(tiDomain3 d) {
    return promote().difference(d);
  }
  public boolean isEqual(tiDomain3 d) {
    return promote().isEqual(d);
  }
  public boolean isNotEqual(tiDomain3 d) {
    return promote().isNotEqual(d);
  }
  public boolean isSubset(tiDomain3 d) {
    return promote().isSubset(d);
  }
  public boolean isStrictSubset(tiDomain3 d) {
    return promote().isStrictSubset(d);
  }
  public boolean isSuperset(tiDomain3 d) {
    return promote().isSuperset(d);
  }
  public boolean isStrictSuperset(tiDomain3 d) {
    return promote().isStrictSuperset(d);
  }

  // RectDomain Resizing
  private tiRectDomain3 resize(int k, int dir, int s) {
    int direction;
    Point<3> new_p0, new_p1, new_stride;

    new_p0 = p0;
    new_p1 = p1;
    new_stride = stride;
    direction = tiDomainUtils.direction_value(dir);

    if ((s == stride.get(direction)) || (stride.get(direction) == 0)) {

      if (tiDomainUtils.direction_backward(dir)) {
	new_p0 = p0.set(direction, p0.get(direction) - k*s);
      } else {
	new_p1 = p1.set(direction, p1.get(direction) + k*s);
      }
      new_stride = stride.set(direction, s);

      return new tiRectDomain3(new_p0, new_p1, new_stride);
    } else {
      // !!! error - result not a Rect domain
      System.out.println("Runtime: (Error) not a RectDomain in accrete/shrink.");
      return new tiRectDomain3();
    }
  }

  public tiRectDomain3 accrete(int k, int dir, int s) {
    return resize(k, dir, s);
  }
  public tiRectDomain3 accrete(int k, int dir) {
    return resize(k, dir, 1);
  }
  public tiRectDomain3 accrete(int k, Point<3> p) {
    tiRectDomain3 result = this;
    for (int i = 1; i <= p.arity(); i++) {
      result = result.resize(k, i, p.get(tiDomainUtils.direction_value(i)));
      result = result.resize(k, -i, p.get(tiDomainUtils.direction_value(-i)));
    }
    return result;
  }
  public tiRectDomain3 accrete(int k) {
    return accrete(k, Point<3>.all(1));
  }

  public tiRectDomain3 shrink(int k, int dir) {
    return resize(-k,
		  dir,
		  stride.get(tiDomainUtils.direction_value(dir)));
  }
  public tiRectDomain3 shrink(int k) {
    tiRectDomain3 result = this;
    for (int i = 1; i <= getArity(); i++) {
      result = result.resize(-k,
			     i,
			     stride.get(tiDomainUtils.direction_value(i)));
      result = result.resize(-k,
			     -i,
			     stride.get(tiDomainUtils.direction_value(-i)));
    }
    return result;
  }
  public tiRectDomain3 border(int k, int dir, int s) {
    int direction;
    Point<3> new_p0, new_p1, new_stride;

    new_p0 = p0;
    new_p1 = p1;
    new_stride = Point<3>.all(1);
    direction = tiDomainUtils.direction_value(dir);
    if (tiDomainUtils.direction_backward(dir)) {
      new_p1 = new_p1.set(direction, new_p0.get(direction) - s + k);
      new_p0 = new_p0.set(direction, new_p0.get(direction) - s);
    } else {
      new_p0 = new_p0.set(direction, new_p1.get(direction) + s - k);
      new_p1 = new_p1.set(direction, new_p1.get(direction) + s);
    }

    return new tiRectDomain3(new_p0, new_p1, new_stride);
  }
  public tiRectDomain3 border(int k, int dir) {
    return border(k, dir, 1);
  }
  public tiRectDomain3 border(int dir) {
    return border(1, dir, 1);
  }

  // Point Relations
  public tiRectDomain3 add(Point<3> p) {
    // this + p
    return new tiRectDomain3(p0.add(p), p1.add(p), stride);
  }
  public tiRectDomain3 difference(Point<3> p) {
    // this - p
    return new tiRectDomain3(p0.sub(p), p1.sub(p), stride);
  }
  public tiRectDomain3 multiply(Point<3> p) {
    // this * p
    Point<3> point_one = Point<3>.all(1);
    Point<3> new_p0 = p0.mul(p);
    Point<3> new_p1 = p1.sub(point_one).mul(p);
    Point<3> new_stride = stride.mul(p);

    for (int i = 0; i < getArity(); i++) {
      if (p.get(i) < 0) {
	int saved_p0 = new_p0.get(i);
	new_p0 = new_p0.setThis(i, new_p1.get(i));
	new_p1 = new_p1.setThis(i, saved_p0);
	new_stride = new_stride.setThis(i, -new_stride.get(i));
      }
    }
    new_p1 = new_p1.add(point_one);

    return new tiRectDomain3(new_p0, new_p1, new_stride);
  }
  public tiRectDomain3 divide(Point<3> p) {
    // this / p
    Point<3> point_one = Point<3>.all(1);
    Point<3> new_p0 = p0.div(p);
    Point<3> new_p1 = p1.sub(point_one).div(p);
    Point<3> new_stride = stride.div(p);

    for (int i = 0; i < getArity(); i++) {
      if (stride.get(i) < p.get(i)) {
	new_stride = new_stride.setThis(i, 1);
      }
      if (p.get(i) < 0) {
	int saved_p0 = new_p0.get(i);
	new_p0 = new_p0.setThis(i, new_p1.get(i));
	new_p1 = new_p1.setThis(i, saved_p0);
	new_stride = new_stride.setThis(i, -new_stride.get(i));
      }
    }
    new_p1 = new_p1.add(point_one);

    return new tiRectDomain3(new_p0, new_p1, new_stride);
  }
  public tiRectDomain3 permute(Point<3> p) {
    return new tiRectDomain3(p0.permute(p),
				   p1.permute(p),
				   stride.permute(p));
  }
  public tiRectDomain3 translate(Point<3> p) {
    // this + p
    return new tiRectDomain3(p0.add(p), p1.add(p), stride);
  }
  public tiRectDomain3 inject(Point<3> p) {
    // this * p
    return new tiRectDomain3(p0.mul(p),
				   p1.mul(p),
				   stride.mul(p));
  }
  public tiRectDomain3 project(Point<3> p) {
    // this / p
    return new tiRectDomain3(p0.div(p),
				   p1.div(p),
				   stride.div(p));
  }

  // Shape information
  public int getArity() {
    return 3;
  }
  public boolean doesContain(Point<3> p) {
    // this contains p
    if ((p.isGreaterThanEqual(p0)) && (p.isLessThan(p1))) {
      for (int i = 0; i < getArity(); i++) {
	if (stride.get(i) == 0) {
	  if (p.get(i) != p0.get(i)) {
	    return false;
	  }
	} else {
	  if (((p.get(i) - p0.get(i)) % stride.get(i)) != 0) {
	    return false;
	  }
	}
      }
      return true;
    }
    return false;
  }
  public boolean isRectangular() {
    return true;
  }
  public int getNumPoints() {
    int side_mult = 1;
    for (int i = 0; i < getArity(); i++) {
      side_mult = side_mult * getNumSidePoints(i);
    }
    return side_mult;
  }
  public Point<3> getLowerBound() {
    return p0;
  }
  public Point<3> getUpperBound() {
    return p1;
  }
  public Point<3> getUpperMaxPoint() {
    return p1.sub(Point<3>.all(1));
  }
  public Point<3> getStride() {
    return stride;
  }
  public Point<3> getLoopingStride() {
    return loopStride;
  }
  public tiRectDomain3 getBoundingBox() {
    return this;
  }
  public tiRectDomainList3 getRectangles() {
    tiRectDomainList3 rdl = new (tiDomains.region) tiRectDomainList3();
    rdl.append(this);
    return rdl;
  }
  public int getNumSidePoints(int dir) {
    if (stride.get(dir) == 0) {
      if (p1.get(dir) <= p0.get(dir)) {
	return 0;
      } else {
	return 1;
      }
    } else {
      return ((p1.get(dir) - 1 - p0.get(dir)) / stride.get(dir)) + 1;
    }
  }

  // Dynamic optimizations
  private tiDomain3 promote() {
    // Promote the object to be represented in a more general type.
    // For now, always promote to a MultiRectDomainA.
    if (isNull()) {
      return new (tiDomains.region) tiMultiRectADomain3();
    } else {
      return new (tiDomains.region) tiMultiRectADomain3(this);
    }
  }

  // Convert to a string
  public String local toString() {
    return "["
	+ getLowerBound().toString()
	+ ":"
	+ getUpperMaxPoint().toString()
	+ ":"
	+ getLoopingStride().toString()
	+ "]";
  }

  // Debugging
  /*
  public void print(String desc) {
    Point<3> p = getLowerBound();
    System.out.print(desc);
    System.out.print("[[");
    for (int i = 0; i < p.arity(); i++) {
      if (i != 0) System.out.print(",");
      System.out.print(p.get(i));
    }
    System.out.print("] : ");
    Point<3> p2 = getUpperBound();
    System.out.print("[");
    for (int i = 0; i < p2.arity(); i++) {
      if (i != 0) System.out.print(",");
      System.out.print(p2.get(i));
    }
    System.out.print("] : ");
    Point<3> p3 = getStride();
    System.out.print("[");
    for (int i = 0; i < p3.arity(); i++) {
      if (i != 0) System.out.print(",");
      System.out.print(p3.get(i));
    }
    System.out.println("]]");
  }
  */

  // State
  private Point<3> p0;
  private Point<3> p1;
  private Point<3> stride;
  private Point<3> loopStride;

  // debugging helpers
  
  private boolean trivialStride() {
    for (int dim = 1; dim <= 3; ++dim)
      if (loopStride[dim] != 1)
	return false;
    return true;
  }

  private void println() {
    System.out.println(toString());
  }
}


// Local variables:
// c-basic-offset: 2
// End:
