package ti.lang;
import java.io.IOException;


final public class NativeUtils
{
    // These utility methods are used by native code.  As much of the
    // logic as possible has been pulled out of the native code so
    // that local qualification inference can see what is going on.

    // Build a string out of an array of Unicode characters.  This is
    // used by native utility function java_string_build_16().

    public static String local buildString( char[] local chars )
    {
	return new String( chars );
    }

    
    // Build a string out of an array of ASCII bytes.  Used by native
    // utility function java_string_build_8().

    public static String local buildString( byte[] local bytes )
    {
	return new String( bytes, 0 );
    }


    // Extract an ASCII byte array out of a string.  Used by native
    // utility functions jstring2Cstring(), stringToDouble(), and
    // jopen().

    public static byte[] local getBytes( String string )
    {
	int length = string.length();
	byte[] local result = new byte[ length ];
	
	string.getBytes( 0, length, result, 0 );
	return result;
    }
    

    // Intern a local string.  Used by literal string table
    // initialization code.
    
    public static String local intern( String local string )
    {
	return (String local) string.intern();
    }


    // Check array bounds.  Throw an array bounds exception if the
    // proposed offset strays outside of safe bounds.  Used by native
    // utility macros JAVA_ARRAY_CHECK_GLOBAL() and
    // JAVA_ARRAY_CHECK_LOCAL().

    public static void checkArrayBounds( int trustedMax, int offset, int length )
    {
	if (offset < 0 || length < 0 || offset + length > trustedMax)
	    throw new ArrayIndexOutOfBoundsException();
    }


    // Throw an illegal argument exception.  Used by native method
    // java.lang.UNUXProcess.forkAndExec().

    public static void throwIllegalArgumentException() throws IllegalArgumentException
    {
	throw new IllegalArgumentException();
    }


    // Throw an I/O exception with the given message.  Used by native
    // utility function tossIOException().

    public static void throwIOException( String local message ) throws IOException
    {
	throw new IOException( message );
    }


    // Throw an out-of-memory error.  Used by native method
    // java.lang.UNIXProcess.forkAndExev().
    
    public static void throwOutOfMemoryError() throws OutOfMemoryError
    {
	throw new OutOfMemoryError();
    }
    

    // Throw a not-cloneable exception.  Use by native utility
    // function allocateClone().

    public static void throwCloneNotSupportedException() throws CloneNotSupportedException
    {
        // !! In Java, the exception has a message -- the class name.
 	throw new CloneNotSupportedException();
    }
    
    
    // Throw a class cast exception.  Used by native utility
    // functions castToClass() and castToInterface().  Also
    // used by code generated for certain cast expressions.

    public static void throwClassCastException() throws ClassCastException
    {
	// !! In Java, the exception has a message -- the name of the
	// dynamic type of the instance being cast.
	
 	throw new ClassCastException();
    }
    
    
    // Throw a number format exception.  Used by native utility macros
    // CHECK_NULL_GLOBAL() and CHECK_NULL_LOCAL().

    public static void throwNullPointerException() throws NullPointerException
    {
	throw new NullPointerException();
    }


    // Throw a number format exception.  Used by native utility
    // function stringToDouble().

    public static void throwNumberFormatException( String message ) throws NumberFormatException
    {
	throw new NumberFormatException( message );
    }


    // Add a property to a property table.  Used by native utility
    // function addProperty().

    public static void addProperty( java.util.Properties props,
				    String local key,
				    String local value )
    {
	props.put( key, value );
    }
}
